# Generated with SMOP  0.41
from libsmop import *
# solar_zenith.m

    
@function
def solar_zenith(lat=None,d=None,t=None,*args,**kwargs):
    varargin = solar_zenith.varargin
    nargin = solar_zenith.nargin

    #function to calculate the cosine of the zenith angle based on the time of day, the
#latitude of the site and the time of year. t is in hours ranging from 0 to
#24.  d is the calender day with J = 1 at January 1. LC is the longitude
#correction (+7.5 to -7.5 either side of standard meridian).
    
    #Solar declination ranges from +23.45 at summer soltice to -23.45 at winter
#solstice and is calculated as:
    
    #make sure to change latitudinal correction for whatever city you are in.
    LC=- 49 / 60
# solar_zenith.m:12
    
    
    ET=solar_ET(d)
# solar_zenith.m:14
    to=12 - LC - ET
# solar_zenith.m:16
    dec=solar_dec(d)
# solar_zenith.m:18
    
    cos_z=(multiply(sind(lat),sind(dec))) + (multiply(multiply(cosd(lat),cosd(dec)),cosd(dot(15.0,(t - to)))))
# solar_zenith.m:20
    zen=acosd(cos_z)
# solar_zenith.m:22
    bad=find(zen > 90)
# solar_zenith.m:24
    zen[bad]=90
# solar_zenith.m:25
    #Written January 2007 by Natasha Kenny
    


# -*- coding: utf-8 -*-
"""
This code is a translation of radiation related function in the COMFA model

Translated on Fri Oct 23 
@author: xlinfr
"""

import numpy as np
from ....util.SEBESOLWEIGCommonFiles import Solweig_v2015_metdata_noload as metload
from ....util.SEBESOLWEIGCommonFiles.diffusefraction import diffusefraction


def CNRRabs_Total(alpha,L,D,Lin,Lup,Kin,Kup,Atr,d,t,lat,A,emis,Aeff):
    # VERTICAL CYLINDER MODEL 
    # Used to calculate the total radiation absorbed by a cylinder, such as the CRT, with inputs
    # alpha (albedo of cylinder), Kin (incoming shortwave measured by CNR net
    # radiometer, W/m2), Kup (reflected shortwave measured by CNR, W/m2), Lin
    # (CNR incoming longwave radiation, W/m2), Lup (CNR1 outgoing longwave
    # radition, W/m2), L (length of cylinder, m), D (diamter of cylinder, m).
    # 
    #Also need solar zenith, which is calculated from: lat (latitude, degrees), d (days, Jan 1=1), t (decimal
    # time, 24h decimal time).
        
    #Atr - atmospheric transmittance - can calculate for the day/time of day or
    #estimate based on sky conditions.
        
        # LOAD data ....
        
        #set constants
        # D=0.01
    # CNRRabs_Total.m:20
        
        # L=0.1
    # CNRRabs_Total.m:21
        
        # alpha=0.37
    # CNRRabs_Total.m:22
        
        #     E.g., 
    #     Kin = metdata(:,10); #put in column for Kin from metdata file
    #     Kup = metdata(:,11); 
    #     Lin  = metdata(:,20);
    #     Lup = metdata(:,21); 
    #     d = metdata(:,24);
    #     Atr = 0.70; 
    #     lat = 33.75;
    #     A = 992;
    
    # Aeff=0.78 # Effective area of body. 0.78 for standing from Campebll and Normal (1998) (0.70 for sitting) Moved to main function

    Kin_abs, zen, Kd = CNR_Kinabs_meas(alpha, Kin, L, D, A, lat, d, t, Atr)
    Kup_abs = CNR_Kup(alpha, Kup, L, D)
    Lin_abs = LinMeas_abs(L, D, Lin, emis)
    Lup_abs = LupMeas_abs(L, D, Lup, emis)
    Acyl = CRT_Acyl(L,D)
    Total_CNRRabs = np.dot(Aeff,((Kin_abs + Kup_abs + Lin_abs + Lup_abs) / (Acyl)))

    return Total_CNRRabs, zen, Kd

def COMFA_RAD_SPATIAL_TC(L, D, Lin, Lup, Kin, Kup, emis, alpha, Aeff, Kd, metdata, location, utc):
    """
    Same as CNRRabs_Total but using a SPN1 sensor to obtain Kd or
    Same as CNRRabs_Total but using reindl et al for Kd and sun_position.py
    """

    deg2rad = np.pi/180.

    # location = {'longitude': lon, 'latitude': lat, 'altitude': alt}
    YYYY, altitude, azimuth, zen, jday, leafon, dectime, altmax = metload.Solweig_2015a_metdata_noload(metdata, location, utc)

    Kb = Kin - Kd
    # Not perpendicular but horrisontal surface in COMFA Kb = (Kin - Kd)/(np.sin(altitude*deg2rad))

    Acs = CRT_Acs(L,D)

    if altitude > 0:
        # Kdirect perpendicular to the beam
        Kp = Kb/np.cos(zen)
        Kb_abs = (1 - alpha) * (Kp * np.sin(zen)) * Acs
    else:
        Kb_abs = 0
        Kd = 0
    # Kb_abs = np.dot(np.dot((1 - alpha),(np.dot(Kb,np.sin(zen)))),Acs)
    Acyl = CRT_Acyl(L,D)
    # Kd_abs = np.dot(np.dot(np.dot((1 - alpha),Kd),0.5),Acyl)
    Kd_abs = (1 - alpha) * Kd * 0.5 * Acyl

    Kin_abs = (Kb_abs + Kd_abs)

    # same as original
    Kup_abs = CNR_Kup(alpha, Kup, L, D)
    Lin_abs = LinMeas_abs(L, D, Lin, emis)
    Lup_abs = LupMeas_abs(L, D, Lup, emis)
    Acyl = CRT_Acyl(L,D)
    # Total_CNRRabs_solweig = np.dot(Aeff,((Kin_abs + Kup_abs + Lin_abs + Lup_abs) / (Acyl)))
    Total_CNRRabs_solweig = Aeff * ((Kin_abs + Kup_abs + Lin_abs + Lup_abs) / Acyl)

    return Total_CNRRabs_solweig

def Rad_Total_solweig(L,D,Lin,Lup,Kin,Kup,yyyy,hour,minu,doy,lat,lon,alt,emis,alpha,Aeff,utc, Kd=None, Ta=None, RH=None):
    """
    Same as CNRRabs_Total but using a SPN1 sensor to obtain Kd or
    Same as CNRRabs_Total but using reindl et al for Kd and sun_position.py
    """

    deg2rad = np.pi/180.
    metdata = np.zeros((1, 24)) - 999. #TODO: Move out of function
    # doy = day_of_year(yyyy, month, day)
    metdata[0, 0] = yyyy
    metdata[0, 1] = doy
    metdata[0, 2] = hour
    metdata[0, 3] = minu
    # metdata[0, 11] = Ta
    # metdata[0, 10] = RH
    # metdata[0, 14] = Kin
    # metdata[0, 21] = Kd

    location = {'longitude': lon, 'latitude': lat, 'altitude': alt}
    YYYY, altitude, azimuth, zen, jday, leafon, dectime, altmax = metload.Solweig_2015a_metdata_noload(metdata, location, utc)

    if Kd == None: # Kd and Kb is estimated with Reindl et al. 1990
        Itoa = 1370.0  # Effective solar constant
        D_ = sun_distance(doy)
        I0et = Itoa * np.cos(zen) * D_  # extra terrestial solar radiation
        Kt = Kin / I0et
        _, Kd = diffusefraction(Kin, altitude, Kt, Ta, RH)
        Kb = Kin - Kd
    else:
        Kb = Kin - Kd
        # Not perpendicular but horrisontal surface in COMFA Kb = (Kin - Kd)/(np.sin(altitude*deg2rad))

    Acs = CRT_Acs(L,D)

    if altitude > 0:
        # Kdirect perpendicular to the beam
        Kp = Kb/np.cos(zen)
        Kb_abs = (1 - alpha) * (Kp * np.sin(zen)) * Acs
    else:
        Kb_abs = 0
        Kd = 0
    # Kb_abs = np.dot(np.dot((1 - alpha),(np.dot(Kb,np.sin(zen)))),Acs)
    Acyl = CRT_Acyl(L,D)
    # Kd_abs = np.dot(np.dot(np.dot((1 - alpha),Kd),0.5),Acyl)
    Kd_abs = (1 - alpha) * Kd * 0.5 * Acyl

    Kin_abs = (Kb_abs + Kd_abs)

    # same as original
    Kup_abs = CNR_Kup(alpha, Kup, L, D)
    Lin_abs = LinMeas_abs(L, D, Lin, emis)
    Lup_abs = LupMeas_abs(L, D, Lup, emis)
    Acyl = CRT_Acyl(L,D)
    # Total_CNRRabs_solweig = np.dot(Aeff,((Kin_abs + Kup_abs + Lin_abs + Lup_abs) / (Acyl)))
    Total_CNRRabs_solweig = Aeff * ((Kin_abs + Kup_abs + Lin_abs + Lup_abs) / Acyl)

    return Total_CNRRabs_solweig, zen/(np.pi/180), Kd


def CNR_Kinabs_meas(alpha,Kin,L,D,A,lat,d,t,Atr):
    #VERTICAL CYLINDER MODEL USING THE PERPENDICULAR DIRECT IRRADIANCE AND
    #DIFFUSE BEAMS SEPARATELY
    #Used to determine the total incoming solar radiation (K) absorbed by the
    #human using the measured CNR1 net radiometer readings (W/m2), with inputs alpha (albedo of the cylinder), Kin (incoming solar
    #radiation measured by the net radiometer), L (length of the cylinder, cm),
    #D (diameter of the cylinder, cm), A is altitude in m, lat (latitude,
    #degrees), d (day, Jan 1 =1), and t (time, in decimal 24 hours)
    
    deg2rad = np.pi/180.
    Kb=Ratio_Kb(Kin,A,lat,d,t,Atr)
    Kd= Kin - Kb
    zen=solar_zenith(lat,d,t)
    Acs=CRT_Acs(L,D)
    Kb_abs=np.dot(np.dot((1 - alpha),(np.dot(Kb,np.sin(zen*deg2rad)))),Acs)
    Acyl=CRT_Acyl(L,D)
    Kd_abs=np.dot(np.dot(np.dot((1 - alpha),Kd),0.5),Acyl)
    Kin_abs=(Kb_abs + Kd_abs)

    return Kin_abs, zen, Kd
    
    # Written August, 2006 by Natasha Kenny
    # Revised January, 2007 to break up the diffuse and beam radiation
    # updated Jan 2016 Aaron Hardin & Jenni Vanos with explanations from Terry
    # Gillespie for ratio model

def Ratio_Kb(Kin,A,lat,d,t,Atr):
    # function is used to estimate incoming shortwave diffuse radiation under
    # clear sky conditions with inputs A(alititude, m), lat(latitude, degrees),
    # d (days, Jan 1 =1), t (time, 24 hours)
    
    m=opt_m(A,lat,d,t)
    Kb=Kin / ((1 + (np.dot(0.3,(1 - Atr ** m)))) / (Atr ** m))

    return Kb

def opt_m(A,lat,d,t):
    #optical air mass number used for input into the equations which estimate
    #incoming shortwave radiation under clear sky conditions
    deg2rad = np.pi/180.
    Pa=atm_P(A)
    zen=solar_zenith(lat,d,t)
    m=Pa / (np.dot(101.3,np.cos(zen*deg2rad)))

    # written January/07 by Natasha Kenny

    return m


def atm_P(A):

    # Written by Jenni Vanos, Aug 2009 for use in COMFA CNR to find Rabs.
    
    Po=101.3 #TODO move out? Pressure? 
    Pa=np.dot(Po,np.exp(- A / 8200))
    
    return Pa


def solar_zenith(lat,d,t):
    #function to calculate the cosine of the zenith angle based on the time of day, the
    #latitude of the site and the time of year. t is in hours ranging from 0 to
    #24.  d is the calender day with J = 1 at January 1. LC is the longitude
    #correction (+7.5 to -7.5 either side of standard meridian).
        
    #Solar declination ranges from +23.45 at summer soltice to -23.45 at winter
    #s olstice and is calculated as:
    deg2rad = np.pi/180.
    #make sure to change latitudinal correction for whatever city you are in.
    
    LC=-49/60. #TODO Remove since is should vary with location. What us this?

    ET=solar_ET(d)
    to=12 - LC - ET
    dec=solar_dec(d)
    
    cos_z=(np.dot(np.sin(lat*deg2rad),np.sin(dec*deg2rad))) + (np.dot(np.dot(np.cos(lat*deg2rad),np.cos(dec*deg2rad)),np.cos(np.dot(15.0,(t - to))*deg2rad)))
    zen=np.arccos(cos_z)*(180/np.pi)
    zen[zen>90]=90.

    # bad=np.where(zen > 90)
    # zen[bad]=90

    #Written January 2007 by Natasha Kenny

    return zen


def solar_ET(d):
    # used to calculate the equation of time for use in calcuating the zenith
    # angle of the sun.
    deg2rad = np.pi/180.
    f=279.575 + np.dot(0.9856,d)
    ET=(np.dot(- 104.7,np.sin(f*deg2rad)) + np.dot(596.2,np.sin(np.dot(2.0,f)*deg2rad)) + np.dot(4.3,np.sin(np.dot(3.0,f)*deg2rad)) - np.dot(12.7,np.sin(np.dot(4.0,f)*deg2rad)) - np.dot(429.3,np.cos(f*deg2rad)) - np.dot(2.0,np.cos(np.dot(2.0,f)*deg2rad)) + np.dot(19.3,np.cos(np.dot(3.0,f)*deg2rad))) / 3600.

    return ET


def solar_dec(d):
    # written to calcuate the solar declination with ranges from +23.45 at
    # summer solstice to -23.45 at the winter solstice.  d is the calendar day
    # with d = 1 being January 1st.
    
    #sin_dec = 0.39785 .* sind(278.97 + 0.9856 .* d + 1.9165 .* sind(356.6 +
    #0.9856 .* d)); from Campbell and norman pg. 168
    
    #dec = asind(sin_dec);
    deg2rad = np.pi/180.
    dec=np.dot(- 23.4,np.cos((np.dot(360,(d + 10)) / 365)*deg2rad))

    return dec


def CRT_Acs(L,D):
    # vertical cross sectional area of the cylinder in m
    
    # Written by Jenni Vanos, Aug 2009, to find the XC area of a cylinder,
    # where L = length (m) and D = diameter (m)- #for CRT L = ~0.10m and D =
    # 0.01m - ratio is most important part to keep constant for changing
    # cylidner size. 

    #Used in Kb_abs = (1-alpha).* (Kb .* sind(zen)).* Acs;
    
    Acs=np.dot(D,L)

    return Acs


def CNR_Kup(alpha,Kup,L,D):
    # Used to calculate the total reflected solar radiation (K) absorbed by the
    # a cylinder with inputs alpha (cylinder albedo), Kup (measured reflected solar
    # radiation by the CNR net radiometer (W/m2), Acyl (area of the cylinder, m2)
    
    Acyl=CRT_Acyl(L,D)
    Kup_abs=np.dot(np.dot(np.dot((1 - alpha),Kup),0.5),Acyl)

    return Kup_abs


def CRT_Acyl(L,D):
    #Function written August 2009 by Jenni Vanos for calculating the Area of
    #the CRT for use in the CNR Comfa model for finidng Rabs.
    
    #L = length (m) - #for CRT this is ~0.10m
    #D = diamater (m) - #for CRT this is ~0.01m
    
    # pi=3.14
    Acyl=np.dot(np.dot(2,np.pi),(D / 2) ** 2) + np.dot(np.dot(np.dot(2.0,np.pi),(D / 2)),L)

    return Acyl


def LinMeas_abs(L,D,Lin,e):
    #Used to Calculate the incoming longwave radiation absorbed by the cylinder with
    #inputs epsilon (emissivity of the cylinder, 0.95), Ta (air temp, degreesC)
    #L (length of cylinder, cm), D (diameter of cylinder, m)
    
    epsilon= e #0.95 moved to main function
    Acyl=CRT_Acyl(L,D)
    LinMeas_abs=np.dot(np.dot(np.dot(epsilon,Lin),0.5),Acyl)

    return LinMeas_abs


def LupMeas_abs(L, D, Lup, e):

    #Used to calculate the estimated outgoing terrestrial longwave absorbed by the
    #cylinder  for use in the Rabs (Estimation) Model with inputs Ta (air temperature),
    #L (length of cylinder, m), D (diameter of cylinder, cm)
    
    Acyl = CRT_Acyl(L,D)
    epsilon = e # 0.95 #mover to main function
    LupMeas_abs = np.dot(np.dot(np.dot(epsilon,Lup),0.5),Acyl)

    return LupMeas_abs

def day_of_year(yyyy, month, day):
        if (yyyy % 4) == 0:
            if (yyyy % 100) == 0:
                if (yyyy % 400) == 0:
                    leapyear = 1
                else:
                    leapyear = 0
            else:
                leapyear = 1
        else:
            leapyear = 0

        if leapyear == 1:
            dayspermonth = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
        else:
            dayspermonth = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

        doy = np.sum(dayspermonth[0:month - 1]) + day

        return doy

def sun_distance(jday):
    """
    #% Calculatesrelative earth sun distance
    #% with day of year as input.
    #% Partridge and Platt, 1975
    """
    b = 2.*np.pi*jday/365.
    D = np.sqrt((1.00011+np.dot(0.034221, np.cos(b))+np.dot(0.001280, np.sin(b))+np.dot(0.000719,
                                        np.cos((2.*b)))+np.dot(0.000077, np.sin((2.*b)))))
    return D


"""Adjusted code by Fredrik Lindberg, orginally from Dag Wastberg, Rambolls"""

import numpy as np
from collections import namedtuple

PET_person=namedtuple("PET_person","mbody age height activity sex clo")


def calculate_PET_index(Ta, Pa, Tmrt, va, pet):
    mbody=pet.mbody
    age=pet.age
    height=pet.height
    activity=pet.activity
    sex=pet.sex
    clo=pet.clo
    pet_index=np.zeros_like(Tmrt)
    for y in range(pet_index.shape[0]):
        for x in range(pet_index.shape[1]):
            pet_index[y,y]=_PET(Ta[x],Pa[x],Tmrt[x][y],va[x][y],mbody,age,height,activity,clo,sex)

def calculate_PET_grid(Ta, RH, Tmrt, va, pet, feedback):
    mbody=pet.mbody
    age=pet.age
    height=pet.height
    activity=pet.activity
    sex=pet.sex
    clo=pet.clo
    pet_index=np.zeros_like(Tmrt)
    total = 100. / (int(pet_index.shape[0] * pet_index.shape[1]))
    index = 0
    # print(Tmrt.shape)
    # print(va.shape)
    for y in range(pet_index.shape[0]):
        if feedback.isCanceled():
            feedback.setProgressText("Calculation cancelled")
            break
        for x in range(pet_index.shape[1]):
            if va[y, x] > 0:
                index = index + 1
                feedback.setProgress(int(index * total))
                pet_index[y,x]=_PET(Ta,RH,Tmrt[y,x],va[y,x],mbody,age,height,activity,clo,sex)
            else:
                pet_index[y, x] = -9999

    return pet_index

def calculate_PET_index_vec(Ta, Pa, Tmrt, va,pet):
    mbody=pet.mbody
    age=pet.age
    height=pet.height
    activity=pet.activity
    sex=pet.sex
    clo=pet.clo

    pet_index=_PET(Ta,Pa,Tmrt,va,mbody,age,height,activity,clo,sex)

def _PET(ta,RH,tmrt,v,mbody,age,ht,work,icl,sex):
    """
    Args:
        ta: air temperature
        RH: relative humidity
        tmrt: Mean Radiant temperature
        v: wind at pedestrian heigh
        mbody: body masss (kg)
        age: person's age (years)
        ht: height (meters)
        work: activity level (W)
        icl: clothing amount (0-5)
        sex: 1=male 2=female
    Returns:
    """

    # humidity conversion
    vps = 6.107 * (10. ** (7.5 * ta / (238. + ta)))
    vpa = RH * vps / 100  # water vapour presure, kPa

    po = 1013.25  # Pressure
    p = 1013.25  # Pressure
    rob = 1.06
    cb = 3.64 * 1000
    food = 0
    emsk = 0.99
    emcl = 0.95
    evap = 2.42e6
    sigma = 5.67e-8
    cair = 1.01 * 1000

    eta = 0  # No idea what eta is

    c_1 = 0.
    c_2 = 0.
    c_3 = 0.
    c_4 = 0.
    c_5 = 0.
    c_6 = 0.
    c_7 = 0.
    c_8 = 0.
    c_9 = 0.
    c_10 = 0.
    c_11 = 0.

    # INBODY
    metbf = 3.19 * mbody ** (3 / 4) * (1 + 0.004 * (30 - age) + 0.018 * ((ht * 100 / (mbody ** (1 / 3))) - 42.1))
    metbm = 3.45 * mbody ** (3 / 4) * (1 + 0.004 * (30 - age) + 0.010 * ((ht * 100 / (mbody ** (1 / 3))) - 43.4))
    if sex == 1:
        met = metbm + work
    else:
        met = metbf + work

    h = met * (1 - eta)
    rtv = 1.44e-6 * met

    # sensible respiration energy
    tex = 0.47 * ta + 21.0
    eres = cair * (ta - tex) * rtv

    # latent respiration energy
    vpex = 6.11 * 10 ** (7.45 * tex / (235 + tex))
    erel = 0.623 * evap / p * (vpa - vpex) * rtv
    # sum of the results
    ere = eres + erel

    # calcul constants
    feff = 0.725
    adu = 0.203 * mbody ** 0.425 * ht ** 0.725
    facl = (-2.36 + 173.51 * icl - 100.76 * icl * icl + 19.28 * (icl ** 3)) / 100
    if facl > 1:
        facl = 1
    rcl = (icl / 6.45) / facl
    y = 1

    # should these be else if statements?
    if icl < 2:
        y = (ht-0.2) / ht
    if icl <= 0.6:
        y = 0.5
    if icl <= 0.3:
        y = 0.1

    fcl = 1 + 0.15 * icl
    r2 = adu * (fcl - 1. + facl) / (2 * 3.14 * ht * y)
    r1 = facl * adu / (2 * 3.14 * ht * y)
    di = r2 - r1
    acl = adu * facl + adu * (fcl - 1)

    tcore = [0] * 8

    wetsk = 0
    hc = 2.67 + 6.5 * v ** 0.67
    hc = hc * (p / po) ** 0.55
    c_1 = h + ere
    he = 0.633 * hc / (p * cair)
    fec = 1 / (1 + 0.92 * hc * rcl)
    htcl = 6.28 * ht * y * di / (rcl * np.log(r2 / r1) * acl)
    aeff = adu * feff
    c_2 = adu * rob * cb
    c_5 = 0.0208 * c_2
    c_6 = 0.76075 * c_2
    rdsk = 0.79 * 10 ** 7
    rdcl = 0

    count2 = 0
    j = 1

    while count2 == 0 and j < 7:
        tsk = 34
        count1 = 0
        tcl = (ta + tmrt + tsk) / 3
        count3 = 1
        enbal2 = 0

        while count1 <= 3:
            enbal = 0
            while (enbal*enbal2) >= 0 and count3 < 200:
                enbal2 = enbal
                # 20
                rclo2 = emcl * sigma * ((tcl + 273.2) ** 4 - (tmrt + 273.2) ** 4) * feff
                tsk = 1 / htcl * (hc * (tcl - ta) + rclo2) + tcl

                # radiation balance
                rbare = aeff * (1 - facl) * emsk * sigma * ((tmrt + 273.2) ** 4 - (tsk + 273.2) ** 4)
                rclo = feff * acl * emcl * sigma * ((tmrt + 273.2) ** 4 - (tcl + 273.2) ** 4)
                rsum = rbare + rclo

                # convection
                cbare = hc * (ta - tsk) * adu * (1 - facl)
                cclo = hc * (ta - tcl) * acl
                csum = cbare + cclo

                # core temperature
                c_3 = 18 - 0.5 * tsk
                c_4 = 5.28 * adu * c_3
                c_7 = c_4 - c_6 - tsk * c_5
                c_8 = -c_1 * c_3 - tsk * c_4 + tsk * c_6
                c_9 = c_7 * c_7 - 4. * c_5 * c_8
                c_10 = 5.28 * adu - c_6 - c_5 * tsk
                c_11 = c_10 * c_10 - 4 * c_5 * (c_6 * tsk - c_1 - 5.28 * adu * tsk)
                # tsk[tsk==36]=36.01
                #print(tsk.shape)
                if tsk == 36:
                    tsk = 36.01

                tcore[7] = c_1 / (5.28 * adu + c_2 * 6.3 / 3600) + tsk
                tcore[3] = c_1 / (5.28 * adu + (c_2 * 6.3 / 3600) / (1 + 0.5 * (34 - tsk))) + tsk
                if c_11 >= 0:
                    tcore[6] = (-c_10-c_11 ** 0.5) / (2 * c_5)
                if c_11 >= 0:
                    tcore[1] = (-c_10+c_11 ** 0.5) / (2 * c_5)
                if c_9 >= 0:
                    tcore[2] = (-c_7+abs(c_9) ** 0.5) / (2 * c_5)
                if c_9 >= 0:
                    tcore[5] = (-c_7-abs(c_9) ** 0.5) / (2 * c_5)
                tcore[4] = c_1 / (5.28 * adu + c_2 * 1 / 40) + tsk

                # transpiration
                tbody = 0.1 * tsk + 0.9 * tcore[j]
                sw = 304.94 * (tbody - 36.6) * adu / 3600000
                vpts = 6.11 * 10 ** (7.45 * tsk / (235. + tsk))
                if tbody <= 36.6:
                    sw = 0
                if sex == 2:
                    sw = 0.7 * sw
                eswphy = -sw * evap

                eswpot = he * (vpa - vpts) * adu * evap * fec
                wetsk = eswphy / eswpot
                if wetsk > 1:
                    wetsk = 1
                eswdif = eswphy - eswpot
                if eswdif <= 0:
                    esw = eswpot
                else:
                    esw = eswphy
                if esw > 0:
                    esw = 0

                # diffusion
                ed = evap / (rdsk + rdcl) * adu * (1 - wetsk) * (vpa - vpts)

                # MAX VB
                vb1 = 34 - tsk
                vb2 = tcore[j] - 36.6
                if vb2 < 0:
                    vb2 = 0
                if vb1 < 0:
                    vb1 = 0
                vb = (6.3 + 75 * vb2) / (1 + 0.5 * vb1)

                # energy balance
                enbal = h + ed + ere + esw + csum + rsum + food

                # clothing's temperature
                if count1 == 0:
                    xx = 1
                if count1 == 1:
                    xx = 0.1
                if count1 == 2:
                    xx = 0.01
                if count1 == 3:
                    xx = 0.001
                if enbal > 0:
                    tcl = tcl + xx
                else:
                    tcl = tcl - xx

                count3 = count3 + 1
            count1 = count1 + 1
            enbal2 = 0

        if j == 2 or j == 5:
            if c_9 >= 0:
                if tcore[j] >= 36.6 and tsk <= 34.050:
                    if (j != 4 and vb >= 91) or (j == 4 and vb < 89):
                        pass
                    else:
                        if vb > 90:
                            vb = 90
                        count2 = 1

        if j == 6 or j == 1:
            if c_11 > 0:
                if tcore[j] >= 36.6 and tsk > 33.850:
                    if (j != 4 and vb >= 91) or (j == 4 and vb < 89):
                        pass
                    else:
                        if vb > 90:
                            vb = 90
                        count2 = 1

        if j == 3:
            if tcore[j] < 36.6 and tsk <= 34.000:
                if (j != 4 and vb >= 91) or (j == 4 and vb < 89):
                    pass
                else:
                    if vb > 90:
                        vb = 90
                    count2 = 1

        if j == 7:
            if tcore[j] < 36.6 and tsk > 34.000:
                if (j != 4 and vb >= 91) or (j == 4 and vb < 89):
                    pass
                else:
                    if vb > 90:
                        vb = 90
                    count2 = 1

        if j == 4:
            if (j != 4 and vb >= 91) or (j == 4 and vb < 89):
                pass
            else:
                if vb > 90:
                    vb = 90
                count2 = 1

        j = j + 1

    # PET_cal
    tx = ta
    enbal2 = 0
    count1 = 0
    enbal = 0

    hc = 2.67 + 6.5 * 0.1 ** 0.67
    hc = hc * (p / po) ** 0.55

    while count1 <= 3:
        while (enbal * enbal2) >= 0:
            enbal2 = enbal

            # radiation balance
            rbare = aeff * (1 - facl) * emsk * sigma * ((tx + 273.2) ** 4 - (tsk + 273.2) ** 4)
            rclo = feff * acl * emcl * sigma * ((tx + 273.2) ** 4 - (tcl + 273.2) ** 4)
            rsum = rbare + rclo

            # convection
            cbare = hc * (tx - tsk) * adu * (1 - facl)
            cclo = hc * (tx - tcl) * acl
            csum = cbare + cclo

            # diffusion
            ed = evap / (rdsk + rdcl) * adu * (1 - wetsk) * (12 - vpts)

            # respiration
            tex = 0.47 * tx + 21
            eres = cair * (tx - tex) * rtv
            vpex = 6.11 * 10 ** (7.45 * tex / (235 + tex))
            erel = 0.623 * evap / p * (12 - vpex) * rtv
            ere = eres + erel

            # energy balance
            enbal = h + ed + ere + esw + csum + rsum

            # iteration concerning Tx
            if count1 == 0:
                xx = 1
            if count1 == 1:
                xx = 0.1
            if count1 == 2:
                xx = 0.01
            if count1 == 3:
                xx = 0.001
            if enbal > 0:
                tx = tx - xx
            if enbal < 0:
                tx = tx + xx
        count1 = count1 + 1
        enbal2 = 0

    return tx


"""
/***************************************************************************
 ProcessingUMEP
                                 A QGIS plugin
 UMEP for processing toolbox
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2020-04-02
        copyright            : (C) 2020 by Fredrik Lindberg
        email                : fredrikl@gvc.gu.se
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

"The Shadow generator plugin can be used to generate pixel wise shadow analysis using ground and "
"building digital surface models (DSM). Optionally, vegetation DSMs could also be used. "
"The methodology that is used to generate shadows originates from Ratti and Richens (1990) "
"and is further developed and described in Lindberg and Grimmond (2011).<br>"
"\n"
"------------------<br>"
"Lindberg, F., Grimmond, C.S.B., 2011a. The influence of vegetation and building morphology on shadow patterns and mean radiant temperatures in urban areas: model development and evaluation. Theoret. Appl. Climatol. 105, 311–323 <br>"
"\n"
"Ratti CF, Richens P (1999) Urban texture analysis with image processing techniques. In: Proceedings of the CAADFutures99, Atalanta, GA"
"\n"
"Full manual available via the <b>Help</b>-button."
"https://umep-docs.readthedocs.io/en/latest/processor/Solar%20Radiation%20Daily%20Shadow%20Pattern.html"

"""

__author__ = "Fredrik Lindberg"
__date__ = "2020-04-02"
__copyright__ = "(C) 2020 by Fredrik Lindberg"

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = "$Format:%H$"

import datetime
from pathlib import Path

import pyproj
from rasterio.transform import Affine, xy

from umep import common
from umep.functions import dailyshading as dsh


def generate_shadows(
    dsm_path: str,
    shadow_date_Ymd: str,  # %Y-%m-%d"
    wall_ht_path: str,
    wall_aspect_path: str,
    bbox: list[int, int, int, int],
    out_dir: str,
    shadow_time_HM: int | None = None,  # "%H:%M"
    time_interval_M=30,
    veg_dsm_path: str | None = None,
    trans_veg: float = 3,
    trunk_zone_ht_perc: float = 0.25,
):
    dsm, dsm_transf, dsm_crs, _dsm_nd = common.load_raster(dsm_path, bbox)
    dsm_height, dsm_width = dsm.shape  # y rows by x cols
    dsm_scale = 1 / dsm_transf[1]
    # y is flipped - so return max for lower row
    minx, miny = xy(Affine.from_gdal(*dsm_transf), dsm.shape[0], 0)
    # Define the source and target CRS
    source_crs = pyproj.CRS(dsm_crs)
    target_crs = pyproj.CRS(4326)  # WGS 84
    # Create a transformer object
    transformer = pyproj.Transformer.from_crs(source_crs, target_crs, always_xy=True)
    # Perform the transformation
    lon, lat = transformer.transform(minx, miny)

    # veg transmissivity as percentage
    if not trans_veg >= 0 and trans_veg <= 100:
        raise ValueError("Vegetation transmissivity should be a number between 0 and 100")
    trans = trans_veg / 100.0

    if veg_dsm_path is not None:
        usevegdem = 1
        veg_dsm, veg_dsm_transf, veg_dsm_crs, _veg_dsm_nd = common.load_raster(veg_dsm_path, bbox)
        veg_dsm_height, veg_dsm_width = veg_dsm.shape
        if not (veg_dsm_width == dsm_width) & (veg_dsm_height == dsm_height):
            raise ValueError("Error in Vegetation Canopy DSM: All rasters must be of same extent and resolution")
        trunkratio = trunk_zone_ht_perc / 100.0
        veg_dsm_2 = veg_dsm * trunkratio
        veg_dsm_2_height, veg_dsm_2_width = veg_dsm_2.shape
        if not (veg_dsm_2_width == dsm_width) & (veg_dsm_2_height == dsm_height):
            raise ValueError("Error in Trunk Zone DSM: All rasters must be of same extent and resolution")
    else:
        usevegdem = 0
        veg_dsm = 0
        veg_dsm_2 = 0

    if wall_aspect_path and wall_ht_path:
        print("Facade shadow scheme activated")
        wallsh = 1
        wh_rast, wh_transf, wh_crs, _wh_nd = common.load_raster(wall_ht_path, bbox)
        wh_height, wh_width = wh_rast.shape
        if not (wh_width == dsm_width) & (wh_height == dsm_height):
            raise ValueError("Error in Wall height raster: All rasters must be of same extent and resolution")
        wa_rast, wa_transf, wa_crs, _wa_nd = common.load_raster(wall_aspect_path, bbox)
        wa_height, wa_width = wa_rast.shape
        if not (wa_width == dsm_width) & (wa_height == dsm_height):
            raise ValueError("Error in Wall aspect raster: All rasters must be of same extent and resolution")
    else:
        wallsh = 0
        wh_rast = 0
        wa_height = 0

    dst = 0
    UTC = 0
    target_date = datetime.datetime.strptime(shadow_date_Ymd, "%Y-%m-%d").date()
    year = target_date.year
    month = target_date.month
    day = target_date.day
    if shadow_time_HM is not None:
        onetime = 1
        onetimetime = datetime.datetime.strptime(shadow_time_HM, "%H:%M")
        hour = onetimetime.hour
        minu = onetimetime.minute
        sec = onetimetime.second
    else:
        onetime = 0
        hour = 0
        minu = 0
        sec = 0

    tv = [year, month, day, hour, minu, sec]

    out_path = Path(out_dir)
    out_path.mkdir(parents=True, exist_ok=True)
    out_path_str = str(out_path)

    Path.mkdir(out_path / "facade_shdw_bldgs", parents=True, exist_ok=True)
    Path.mkdir(out_path / "facade_shdw_veg", parents=True, exist_ok=True)
    Path.mkdir(out_path / "shadow_ground", parents=True, exist_ok=True)

    shadowresult = dsh.dailyshading(
        dsm,
        veg_dsm,
        veg_dsm_2,
        dsm_scale,
        lon,
        lat,
        dsm_width,
        dsm_height,
        tv,
        UTC,
        usevegdem,
        time_interval_M,
        onetime,
        out_path_str,
        dsm_transf,
        dsm_crs,
        trans,
        dst,
        wallsh,
        wh_rast,
        wa_rast,
    )

    shfinal = shadowresult["shfinal"]
    common.save_raster(out_path_str + "/shadow_composite.tif", shfinal, dsm_transf, dsm_crs)



import numpy as np

def cylindric_wedge(zen, svfalfa, rows, cols):

    np.seterr(divide='ignore', invalid='ignore')

    # Fraction of sunlit walls based on sun altitude and svf wieghted building angles
    # input: 
    # sun zenith angle "beta"
    # svf related angle "alfa"

    beta=zen
    # alfa=svfalfa
    alfa=np.zeros((rows, cols)) + svfalfa
    # measure the size of the image
    # sizex=size(svfalfa,2)
    # sizey=size(svfalfa,1)
    
    xa=1-2./(np.tan(alfa)*np.tan(beta))
    ha=2./(np.tan(alfa)*np.tan(beta))
    ba=(1./np.tan(alfa))
    hkil=2.*ba*ha
    
    qa=np.zeros((rows, cols))
    # qa(length(svfalfa),length(svfalfa))=0;
    qa[xa<0]=np.tan(beta)/2
    
    Za=np.zeros((rows, cols))
    # Za(length(svfalfa),length(svfalfa))=0;
    Za[xa<0]=((ba[xa<0]**2)-((qa[xa<0]**2)/4))**0.5
    
    phi=np.zeros((rows, cols))
    #phi(length(svfalfa),length(svfalfa))=0;
    phi[xa<0]=np.arctan(Za[xa<0]/qa[xa<0])
    
    A=np.zeros((rows, cols))
    # A(length(svfalfa),length(svfalfa))=0;
    A[xa<0]=(np.sin(phi[xa<0])-phi[xa<0]*np.cos(phi[xa<0]))/(1-np.cos(phi[xa<0]))
    
    ukil=np.zeros((rows, cols))
    # ukil(length(svfalfa),length(svfalfa))=0
    ukil[xa<0]=2*ba[xa<0]*xa[xa<0]*A[xa<0]
    
    Ssurf=hkil+ukil
    
    F_sh=(2*np.pi*ba-Ssurf)/(2*np.pi*ba)#Xa
    
    return F_sh

def cylindric_wedge_voxel(zen, svfalfa):

    np.seterr(divide='ignore', invalid='ignore')

    # Fraction of sunlit walls based on sun altitude and svf wieghted building angles
    # input: 
    # sun zenith angle "beta"
    # svf related angle "alfa"

    beta=zen
    # alfa=svfalfa
    #alfa = voxelTable['svfalfa'].to_numpy()
    # measure the size of the image
    # sizex=size(svfalfa,2)
    # sizey=size(svfalfa,1)
    
    xa=1-2./(np.tan(svfalfa)*np.tan(beta))
    ha=2./(np.tan(svfalfa)*np.tan(beta))
    ba=(1./np.tan(svfalfa))
    hkil=2.*ba*ha
    
    qa=np.zeros((svfalfa.shape[0]))
    # qa(length(svfalfa),length(svfalfa))=0;
    qa[xa<0]=np.tan(beta)/2
    
    Za=np.zeros((svfalfa.shape[0]))
    # Za(length(svfalfa),length(svfalfa))=0;
    Za[xa<0]=((ba[xa<0]**2)-((qa[xa<0]**2)/4))**0.5
    
    phi=np.zeros((svfalfa.shape[0]))
    #phi(length(svfalfa),length(svfalfa))=0;
    phi[xa<0]=np.arctan(Za[xa<0]/qa[xa<0])
    
    A=np.zeros((svfalfa.shape[0]))
    # A(length(svfalfa),length(svfalfa))=0;
    A[xa<0]=(np.sin(phi[xa<0])-phi[xa<0]*np.cos(phi[xa<0]))/(1-np.cos(phi[xa<0]))
    
    ukil=np.zeros((svfalfa.shape[0]))
    # ukil(length(svfalfa),length(svfalfa))=0
    ukil[xa<0]=2*ba[xa<0]*xa[xa<0]*A[xa<0]
    
    Ssurf=hkil+ukil
    
    F_sh=(2*np.pi*ba-Ssurf)/(2*np.pi*ba)#Xa
    
    return F_sh


# Generated with SMOP  0.41
from libsmop import *
# CNR_Kinabs_meas.m

    
@function
def CNR_Kinabs_meas(alpha=None,Kin=None,L=None,D=None,A=None,lat=None,d=None,t=None,Atr=None,*args,**kwargs):
    varargin = CNR_Kinabs_meas.varargin
    nargin = CNR_Kinabs_meas.nargin

    #VERTICAL CYLINDER MODEL USING THE PERPENDICULAR DIRECT IRRADIANCE AND
#DIFFUSE BEAMS SEPARATELY
#Used to determine the total incoming solar radiation (K) absorbed by the
#human using the measured CNR1 net radiometer readings (W/m2), with inputs alpha (albedo of the cylinder), Kin (incoming solar
#radiation measured by the net radiometer), L (length of the cylinder, cm),
#D (diameter of the cylinder, cm), A is altitude in m, lat (latitude,
#degrees), d (day, Jan 1 =1), and t (time, in decimal 24 hours)
    
    Kb=Ratio_Kb(Kin,A,lat,d,t,Atr)
# CNR_Kinabs_meas.m:11
    
    Kd=Kin - Kb
# CNR_Kinabs_meas.m:13
    
    zen=solar_zenith(lat,d,t)
# CNR_Kinabs_meas.m:15
    
    Acs=CRT_Acs(L,D)
# CNR_Kinabs_meas.m:17
    
    Kb_abs=multiply(multiply((1 - alpha),(multiply(Kb,sind(zen)))),Acs)
# CNR_Kinabs_meas.m:19
    
    Acyl=CRT_Acyl(L,D)
# CNR_Kinabs_meas.m:21
    
    Kd_abs=multiply(multiply(multiply((1 - alpha),Kd),0.5),Acyl)
# CNR_Kinabs_meas.m:23
    
    Kin_abs=(Kb_abs + Kd_abs)
# CNR_Kinabs_meas.m:25
    
    # Written August, 2006 by Natasha Kenny
# Revised January, 2007 to break up the diffuse and beam radiation
# updated Jan 2016 Aaron Hardin & Jenni Vanos with explanations from Terry
# Gillespie for ratio model


import json
from types import SimpleNamespace
from typing import Any

import numpy as np

from ... import common
from ...class_configs import EnvironData, ShadowMatrices, SolweigConfig, SvfData, TgMaps, WallsData
from ...util.SEBESOLWEIGCommonFiles.clearnessindex_2013b import clearnessindex_2013b
from . import PET_calculations
from . import Solweig_2025a_calc_forprocessing as so
from . import UTCI_calculations as utci
from .CirclePlotBar import PolarBarPlot
from .patch_characteristics import hemispheric_image
from .wallsAsNetCDF import walls_as_netcdf


def dict_to_namespace(d):
    """Recursively convert dicts to SimpleNamespace."""
    if isinstance(d, dict):
        return SimpleNamespace(**{k: dict_to_namespace(v) for k, v in d.items()})
    elif isinstance(d, list):
        return [dict_to_namespace(i) for i in d]
    else:
        return d


class SolweigRun:
    """Class to run the SOLWEIG algorithm with given configuration."""

    def __init__(self, config: SolweigConfig, params_json_path: str):
        """Initialize the SOLWEIG runner with configuration and parameters."""
        self.config = config
        self.config.validate()
        # Progress tracking settings
        self.progress = None
        self.iters_total: int | None = None
        self.iters_count: int = 0
        # Initialize POI data
        self.poi_names: list[Any] = []
        self.poi_pixel_xys: np.ndarray | None = None
        self.poi_results = []
        # Initialize WOI data
        self.woi_names: list[Any] = []
        self.woi_pixel_xys: np.ndarray | None = None
        self.woi_results = []
        # Load parameters from JSON file
        params_path = common.check_path(params_json_path)
        try:
            with open(params_path) as f:
                params_dict = json.load(f)
                self.params = dict_to_namespace(params_dict)
        except Exception as e:
            raise RuntimeError(f"Failed to load parameters from {params_json_path}: {e}")
        # Prepare core data
        self.dsm_arr: np.ndarray | None = None
        self.scale: float | None = None
        self.rows: int | None = None
        self.cols: int | None = None
        self.location: dict[str, float] | None = None

    def prep_progress(self, num: int) -> None:
        """Prepare progress for environment."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def iter_progress(self) -> bool:
        """Iterate progress ."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def load_epw_weather(self) -> EnvironData:
        """Load weather data from an EPW file."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def load_met_weather(self, header_rows: int = 1, delim: str = " ") -> EnvironData:
        """Load weather data from a MET file."""
        met_path_str = str(common.check_path(self.config.met_path))
        met_data = np.loadtxt(met_path_str, skiprows=header_rows, delimiter=delim)
        return EnvironData(
            self.config,
            self.params,
            YYYY=met_data[:, 0],
            DOY=met_data[:, 1],
            hours=met_data[:, 2],
            minu=met_data[:, 3],
            Ta=met_data[:, 11],
            RH=met_data[:, 10],
            radG=met_data[:, 14],
            radD=met_data[:, 21],
            radI=met_data[:, 22],
            P=met_data[:, 12],
            Ws=met_data[:, 9],
            location=self.location,
            UTC=self.config.utc,
        )

    def load_poi_data(self, trf_arr: list[float]) -> None:
        """Load point of interest (POI) data from a file."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def save_poi_results(self, trf_arr: list[float], crs_wkt: str) -> None:
        """Save results for points of interest (POIs) to files."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def load_woi_data(self, trf_arr: list[float]) -> None:
        """Load wall of interest (WOI) data from a file."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def save_woi_results(self, trf_arr: list[float], crs_wkt: str) -> None:
        """Save results for walls of interest (WOIs) to files."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    def calc_solweig(
        self,
        iter: int,
        buildings: np.ndarray,
        vegdsm: np.ndarray,
        vegdsm2: np.ndarray,
        svfbuveg: np.ndarray,
        bush: np.ndarray,
        lcgrid: np.ndarray,
        wallaspect: np.ndarray,
        wallheight: np.ndarray,
        elvis: float,
        CI: float,
        amaxvalue: float,
        Twater: float,
        first: float,
        second: float,
        firstdaytime: float,
        timeadd: float,
        timestepdec: float,
        posture,
        svf_data: SvfData,
        environ_data: EnvironData,
        tg_maps: TgMaps,
        shadow_mats: ShadowMatrices,
        walls_data: WallsData,
    ):
        """
        Calculate SOLWEIG results for a given iteration.
        Separated from the main run method so that it can be overridden by subclasses.
        Over time we can simplify the function signature by passing consolidated parameters to solweig calc methods.
        """
        return so.Solweig_2025a_calc(
            iter,
            self.dsm_arr,
            self.scale,
            self.rows,
            self.cols,
            svf_data.svf,
            svf_data.svf_north,
            svf_data.svf_west,
            svf_data.svf_east,
            svf_data.svf_south,
            svf_data.svf_veg,
            svf_data.svf_veg_north,
            svf_data.svf_veg_east,
            svf_data.svf_veg_south,
            svf_data.svf_veg_west,
            svf_data.svf_veg_blocks_bldg_sh,
            svf_data.svf_veg_blocks_bldg_sh_east,
            svf_data.svf_veg_blocks_bldg_sh_south,
            svf_data.svf_veg_blocks_bldg_sh_west,
            svf_data.svf_veg_blocks_bldg_sh_north,
            vegdsm,
            vegdsm2,
            self.params.Albedo.Effective.Value.Walls,
            self.params.Tmrt_params.Value.absK,
            self.params.Tmrt_params.Value.absL,
            self.params.Emissivity.Value.Walls,
            posture.Fside,
            posture.Fup,
            posture.Fcyl,
            environ_data.altitude[iter],
            environ_data.azimuth[iter],
            environ_data.zen[iter],
            environ_data.jday[iter],
            self.config.use_veg_dem,
            self.config.only_global,
            buildings,
            self.location,
            environ_data.psi[iter],
            self.config.use_landcover,
            lcgrid,
            environ_data.dectime[iter],
            environ_data.altmax[iter],
            wallaspect,
            wallheight,
            self.config.person_cylinder,
            elvis,
            environ_data.Ta[iter],
            environ_data.RH[iter],
            environ_data.radG[iter],
            environ_data.radD[iter],
            environ_data.radI[iter],
            environ_data.P[iter],
            amaxvalue,
            bush,
            Twater,
            tg_maps.TgK,
            tg_maps.Tstart,
            tg_maps.alb_grid,
            tg_maps.emis_grid,
            tg_maps.TgK_wall,
            tg_maps.Tstart_wall,
            tg_maps.TmaxLST,
            tg_maps.TmaxLST_wall,
            first,
            second,
            svf_data.svfalfa,
            svfbuveg,
            firstdaytime,
            timeadd,
            timestepdec,
            tg_maps.Tgmap1,
            tg_maps.Tgmap1E,
            tg_maps.Tgmap1S,
            tg_maps.Tgmap1W,
            tg_maps.Tgmap1N,
            CI,
            tg_maps.TgOut1,
            shadow_mats.diffsh,
            shadow_mats.shmat,
            shadow_mats.vegshmat,
            shadow_mats.vbshvegshmat,
            self.config.use_aniso,
            shadow_mats.asvf,
            shadow_mats.patch_option,
            walls_data.voxelMaps,
            walls_data.voxelTable,
            environ_data.Ws[iter],
            self.config.use_wall_scheme,
            walls_data.timeStep,
            shadow_mats.steradians,
            walls_data.walls_scheme,
            walls_data.dirwalls_scheme,
        )

    def run(self) -> None:
        """Run the SOLWEIG algorithm."""
        # Load DSM
        self.dsm_arr, dsm_trf_arr, dsm_crs_wkt, dsm_nd_val = common.load_raster(self.config.dsm_path, bbox=None)
        self.scale = 1 / dsm_trf_arr[1]
        self.rows = self.dsm_arr.shape[0]
        self.cols = self.dsm_arr.shape[1]

        left_x = dsm_trf_arr[0]
        top_y = dsm_trf_arr[3]
        lng, lat = common.xy_to_lnglat(dsm_crs_wkt, left_x, top_y)
        alt = np.median(self.dsm_arr)
        if alt < 0:
            alt = 3
        self.location = {"longitude": lng, "latitude": lat, "altitude": alt}

        self.dsm_arr[self.dsm_arr == dsm_nd_val] = 0.0
        if self.dsm_arr.min() < 0:
            dsmraise = np.abs(self.dsm_arr.min())
            self.dsm_arr = self.dsm_arr + dsmraise
        else:
            dsmraise = 0

        # DEM
        # TODO: Is DEM always provided?
        if self.config.dem_path:
            dem_path_str = str(common.check_path(self.config.dem_path))
            dem, _, _, dem_nd_val = common.load_raster(dem_path_str, bbox=None)
            dem[dem == dem_nd_val] = 0.0
            # TODO: Check if this is needed re DSM ramifications
            if dem.min() < 0:
                demraise = np.abs(dem.min())
                dem = dem + demraise

        # Land cover
        if self.config.use_landcover:
            lc_path_str = str(common.check_path(self.config.lc_path))
            lcgrid, _, _, _ = common.load_raster(lc_path_str, bbox=None)
        else:
            lcgrid = None

        # Buildings from land cover option
        # TODO: Check intended logic here
        if not self.config.use_dem_for_buildings and lcgrid is not None:
            # Create building boolean raster from either land cover if no DEM is used
            buildings = np.copy(lcgrid)
            buildings[buildings == 7] = 1
            buildings[buildings == 6] = 1
            buildings[buildings == 5] = 1
            buildings[buildings == 4] = 1
            buildings[buildings == 3] = 1
            buildings[buildings == 2] = 0
        elif self.config.use_dem_for_buildings:
            buildings = self.dsm_arr - dem
            buildings[buildings < 2.0] = 1.0
            buildings[buildings >= 2.0] = 0.0
        else:
            raise ValueError("No DEM or buildings data available.")
        # Save buildings raster if requested
        if self.config.save_buildings:
            common.save_raster(
                self.config.output_dir + "/buildings.tif",
                buildings,
                dsm_trf_arr,
                dsm_crs_wkt,
                dsm_nd_val,
            )

        # Vegetation
        if self.config.use_veg_dem:
            vegdsm, _, _, _ = common.load_raster(self.config.cdsm_path, bbox=None)
            if self.config.tdsm_path:
                vegdsm2, _, _, _ = common.load_raster(self.config.tdsm_path, bbox=None)
            else:
                vegdsm2 = vegdsm * self.params.Tree_settings.Value.Trunk_ratio
        else:
            vegdsm = None
            vegdsm2 = None

        # Load SVF data
        svf_data = SvfData(self.config)

        if self.config.use_veg_dem:
            # amaxvalue
            vegmax = vegdsm.max()
            amaxvalue = self.dsm_arr.max() - self.dsm_arr.min()
            amaxvalue = np.maximum(amaxvalue, vegmax)
            # Elevation vegdsms if buildingDEM includes ground heights
            vegdsm = vegdsm + self.dsm_arr
            vegdsm[vegdsm == self.dsm_arr] = 0
            vegdsm2 = vegdsm2 + self.dsm_arr
            vegdsm2[vegdsm2 == self.dsm_arr] = 0
            # % Bush separation
            bush = np.logical_not(vegdsm2 * vegdsm) * vegdsm
            svfbuveg = svf_data.svf - (1.0 - svf_data.svf_veg) * (
                1.0 - self.params.Tree_settings.Value.Transmissivity
            )  # % major bug fixed 20141203
        else:
            svfbuveg = svf_data.svf
            bush = np.zeros([self.rows, self.cols])
            amaxvalue = 0

        # Load walls
        wallheight, _, _, _ = common.load_raster(self.config.wh_path, bbox=None)
        wallaspect, _, _, _ = common.load_raster(self.config.wa_path, bbox=None)

        # weather data
        if self.config.use_epw_file:
            environ_data = self.load_epw_weather()
        else:
            environ_data = self.load_met_weather(header_rows=1, delim=" ")

        # POIs check
        if self.config.poi_path:
            self.load_poi_data(dsm_trf_arr)

        # Posture settings
        if self.params.Tmrt_params.Value.posture == "Standing":
            posture = self.params.Posture.Standing.Value
        else:
            posture = self.params.Posture.Sitting.Value

        # Radiative surface influence
        first = np.round(posture.height)
        if first == 0.0:
            first = 1.0
        second = np.round(posture.height * 20.0)

        # Import shadow matrices (Anisotropic sky)
        shadow_mats = ShadowMatrices(self.config, self.params, self.rows, self.cols, svf_data)

        # % Ts parameterisation maps
        tg_maps = TgMaps(self.config.use_landcover, lcgrid, self.params, self.rows, self.cols)

        # Import data for wall temperature parameterization
        # Use wall of interest
        if self.config.woi_path:
            self.load_woi_data(dsm_trf_arr)
        walls_data = WallsData(
            self.config,
            self.params,
            self.scale,
            self.rows,
            self.cols,
            environ_data,
            tg_maps,
            self.dsm_arr,
            lcgrid,
        )

        # Initialisation of time related variables
        if environ_data.Ta.__len__() == 1:
            timestepdec = 0
        else:
            timestepdec = environ_data.dectime[1] - environ_data.dectime[0]
        timeadd = 0.0
        firstdaytime = 1.0

        # Save hemispheric image
        if self.config.use_aniso and self.poi_pixel_xys is not None:
            patch_characteristics = hemispheric_image(
                self.poi_pixel_xys,
                shadow_mats.shmat,
                shadow_mats.vegshmat,
                shadow_mats.vbshvegshmat,
                walls_data.voxelMaps,
                walls_data.walls_scheme,
            )

        # Initiate array for I0 values plotting
        if np.unique(environ_data.DOY).shape[0] > 1:
            unique_days = np.unique(environ_data.DOY)
            first_unique_day = environ_data.DOY[unique_days[0] == environ_data.DOY]
            I0_array = np.zeros_like(first_unique_day)
        else:
            first_unique_day = environ_data.DOY.copy()
            I0_array = np.zeros_like(environ_data.DOY)
        # For Tmrt plot
        tmrtplot = np.zeros((self.rows, self.cols))
        # Number of iterations
        num = len(environ_data.Ta)
        # Prepare progress tracking
        self.prep_progress(num)
        # TODO: confirm intent of water temperature handling
        # Assuming it should be initialized to NaN outside the loop so that it can be updated at the start of each day
        Twater = np.nan
        CI = 1.0
        elvis = 0.0
        for i in range(num):
            proceed = self.iter_progress()
            if not proceed:
                break

            # Daily water body temperature - only if land cover is used
            if self.config.use_landcover:  # noqa: SIM102
                # Check if the current time is the start of a new day
                if (environ_data.dectime[i] - np.floor(environ_data.dectime[i])) == 0 or (i == 0):
                    # Find average temperature for the current day
                    Twater = np.mean(environ_data.Ta[environ_data.jday == np.floor(environ_data.dectime[i])])

            # Nocturnal cloudfraction from Offerle et al. 2003
            # Check for start of day
            if (environ_data.dectime[i] - np.floor(environ_data.dectime[i])) == 0:
                # Find all current day idxs
                daylines = np.where(np.floor(environ_data.dectime) == environ_data.dectime[i])
                # np.where returns a tuple, so check the first element
                if len(daylines[0]) > 1:
                    # Get the altitudes for day's idxs
                    alt_day = environ_data.altitude[daylines[0]]
                    # Find all idxs with altitude greater than 1
                    alt2 = np.where(alt_day > 1)
                    # np.where returns a tuple, so check the first element
                    if len(alt2[0]) > 0:
                        # Take the first altitude greater than 1
                        rise = alt2[0][0]
                        # Calculate clearness index for the next time step after sunrise
                        [_, CI, _, _, _] = clearnessindex_2013b(
                            environ_data.zen[i + rise + 1],
                            environ_data.jday[i + rise + 1],
                            environ_data.Ta[i + rise + 1],
                            environ_data.RH[i + rise + 1] / 100.0,
                            environ_data.radG[i + rise + 1],
                            self.location,
                            environ_data.P[i + rise + 1],
                        )
                        if (CI > 1.0) or (~np.isfinite(CI)):
                            CI = 1.0
                    else:
                        CI = 1.0
                else:
                    CI = 1.0
            # Run the SOLWEIG calculations
            (
                Tmrt,
                Kdown,
                Kup,
                Ldown,
                Lup,
                Tg,
                ea,
                esky,
                I0,
                CI,
                shadow,
                firstdaytime,
                timestepdec,
                timeadd,
                tg_maps.Tgmap1,
                tg_maps.Tgmap1E,
                tg_maps.Tgmap1S,
                tg_maps.Tgmap1W,
                tg_maps.Tgmap1N,
                Keast,
                Ksouth,
                Kwest,
                Knorth,
                Least,
                Lsouth,
                Lwest,
                Lnorth,
                KsideI,
                tg_maps.TgOut1,
                TgOut,
                radIout,
                radDout,
                Lside,
                Lsky_patch_characteristics,
                CI_Tg,
                CI_TgG,
                KsideD,
                dRad,
                Kside,
                shadow_mats.steradians,
                voxelTable,
            ) = self.calc_solweig(
                i,
                buildings,
                vegdsm,
                vegdsm2,
                svfbuveg,
                bush,
                lcgrid,
                wallaspect,
                wallheight,
                elvis,
                CI,
                amaxvalue,
                Twater,
                first,
                second,
                firstdaytime,
                timeadd,
                timestepdec,
                posture,
                svf_data,
                environ_data,
                tg_maps,
                shadow_mats,
                walls_data,
            )
            # Save I0 for I0 vs. Kdown output plot to check if UTC is off
            if i < first_unique_day.shape[0]:
                I0_array[i] = I0
            elif i == first_unique_day.shape[0]:
                # Output I0 vs. Kglobal plot
                radG_for_plot = environ_data.radG[first_unique_day[0] == environ_data.DOY]
                dectime_for_plot = environ_data.dectime[first_unique_day[0] == environ_data.DOY]
                fig, ax = plt.subplots()
                ax.plot(dectime_for_plot, I0_array, label="I0")
                ax.plot(dectime_for_plot, radG_for_plot, label="Kglobal")
                ax.set_ylabel("Shortwave radiation [$Wm^{-2}$]")
                ax.set_xlabel("Decimal time")
                ax.set_title("UTC" + str(self.config.utc))
                ax.legend()
                fig.savefig(self.config.output_dir + "/metCheck.png", dpi=150)

            tmrtplot = tmrtplot + Tmrt

            if environ_data.altitude[i] > 0:
                w = "D"
            else:
                w = "N"

            if environ_data.hours[i] < 10:
                XH = "0"
            else:
                XH = ""

            if environ_data.minu[i] < 10:
                XM = "0"
            else:
                XM = ""

            if self.poi_pixel_xys is not None:
                for n in range(0, self.poi_pixel_xys.shape[0]):
                    idx, row_idx, col_idx = self.poi_pixel_xys[n]
                    row_idx = int(row_idx)
                    col_idx = int(col_idx)
                    result_row = {
                        "poi_idx": idx,
                        "yyyy": environ_data.YYYY[i],
                        "id": environ_data.jday[i],
                        "it": environ_data.hours[i],
                        "imin": environ_data.minu[i],
                        "dectime": environ_data.dectime[i],
                        "altitude": environ_data.altitude[i],
                        "azimuth": environ_data.azimuth[i],
                        "kdir": radIout,
                        "kdiff": radDout,
                        "kglobal": environ_data.radG[i],
                        "kdown": Kdown[row_idx, col_idx],
                        "kup": Kup[row_idx, col_idx],
                        "keast": Keast[row_idx, col_idx],
                        "ksouth": Ksouth[row_idx, col_idx],
                        "kwest": Kwest[row_idx, col_idx],
                        "knorth": Knorth[row_idx, col_idx],
                        "ldown": Ldown[row_idx, col_idx],
                        "lup": Lup[row_idx, col_idx],
                        "least": Least[row_idx, col_idx],
                        "lsouth": Lsouth[row_idx, col_idx],
                        "lwest": Lwest[row_idx, col_idx],
                        "lnorth": Lnorth[row_idx, col_idx],
                        "Ta": environ_data.Ta[i],
                        "Tg": TgOut[row_idx, col_idx],
                        "RH": environ_data.RH[i],
                        "Esky": esky,
                        "Tmrt": Tmrt[row_idx, col_idx],
                        "I0": I0,
                        "CI": CI,
                        "Shadow": shadow[row_idx, col_idx],
                        "SVF_b": svf_data.svf[row_idx, col_idx],
                        "SVF_bv": svfbuveg[row_idx, col_idx],
                        "KsideI": KsideI[row_idx, col_idx],
                    }
                    # Recalculating wind speed based on powerlaw
                    WsPET = (1.1 / self.params.Wind_Height.Value.magl) ** 0.2 * environ_data.Ws[i]
                    WsUTCI = (10.0 / self.params.Wind_Height.Value.magl) ** 0.2 * environ_data.Ws[i]
                    resultPET = PET_calculations._PET(
                        environ_data.Ta[i],
                        environ_data.RH[i],
                        Tmrt[row_idx, col_idx],
                        WsPET,
                        self.params.PET_settings.Value.Weight,
                        self.params.PET_settings.Value.Age,
                        self.params.PET_settings.Value.Height,
                        self.params.PET_settings.Value.Activity,
                        self.params.PET_settings.Value.clo,
                        self.params.PET_settings.Value.Sex,
                    )
                    result_row["PET"] = resultPET
                    resultUTCI = utci.utci_calculator(
                        environ_data.Ta[i], environ_data.RH[i], Tmrt[row_idx, col_idx], WsUTCI
                    )
                    result_row["UTCI"] = resultUTCI
                    result_row["CI_Tg"] = CI_Tg
                    result_row["CI_TgG"] = CI_TgG
                    result_row["KsideD"] = KsideD[row_idx, col_idx]
                    result_row["Lside"] = Lside[row_idx, col_idx]
                    result_row["diffDown"] = dRad[row_idx, col_idx]
                    result_row["Kside"] = Kside[row_idx, col_idx]
                    self.poi_results.append(result_row)

            if self.woi_pixel_xys is not None:
                for n in range(0, self.woi_pixel_xys.shape[0]):
                    idx, row_idx, col_idx = self.woi_pixel_xys[n]
                    row_idx = int(row_idx)
                    col_idx = int(col_idx)

                    temp_wall = voxelTable.loc[
                        ((voxelTable["ypos"] == row_idx) & (voxelTable["xpos"] == col_idx)), "wallTemperature"
                    ].to_numpy()
                    K_in = voxelTable.loc[
                        ((voxelTable["ypos"] == row_idx) & (voxelTable["xpos"] == col_idx)), "K_in"
                    ].to_numpy()
                    L_in = voxelTable.loc[
                        ((voxelTable["ypos"] == row_idx) & (voxelTable["xpos"] == col_idx)), "L_in"
                    ].to_numpy()
                    wallShade = voxelTable.loc[
                        ((voxelTable["ypos"] == row_idx) & (voxelTable["xpos"] == col_idx)), "wallShade"
                    ].to_numpy()

                    result_row = {
                        "woi_idx": idx,
                        "woi_name": self.woi_names[idx],
                        "yyyy": environ_data.YYYY[i],
                        "id": environ_data.jday[i],
                        "it": environ_data.hours[i],
                        "imin": environ_data.minu[i],
                        "dectime": environ_data.dectime[i],
                        "Ta": environ_data.Ta[i],
                        "SVF": svf_data.svf[row_idx, col_idx],
                        "Ts": temp_wall,
                        "Kin": K_in,
                        "Lin": L_in,
                        "shade": wallShade,
                        "pixel_x": col_idx,
                        "pixel_y": row_idx,
                    }
                    self.woi_results.append(result_row)

                if self.config.wall_netcdf:
                    netcdf_output = self.config.output_dir + "/walls.nc"
                    walls_as_netcdf(
                        voxelTable,
                        self.rows,
                        self.cols,
                        walls_data.met_for_xarray,
                        i,
                        self.dsm_arr,
                        self.config.dsm_path,
                        netcdf_output,
                    )

            time_code = (
                str(int(environ_data.YYYY[i]))
                + "_"
                + str(int(environ_data.DOY[i]))
                + "_"
                + XH
                + str(int(environ_data.hours[i]))
                + XM
                + str(int(environ_data.minu[i]))
                + w
            )

            if self.config.output_tmrt:
                common.save_raster(
                    self.config.output_dir + "/Tmrt_" + time_code + ".tif",
                    Tmrt,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )
            if self.config.output_kup:
                common.save_raster(
                    self.config.output_dir + "/Kup_" + time_code + ".tif",
                    Kup,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )
            if self.config.output_kdown:
                common.save_raster(
                    self.config.output_dir + "/Kdown_" + time_code + ".tif",
                    Kdown,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )
            if self.config.output_lup:
                common.save_raster(
                    self.config.output_dir + "/Lup_" + time_code + ".tif",
                    Lup,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )
            if self.config.output_ldown:
                common.save_raster(
                    self.config.output_dir + "/Ldown_" + time_code + ".tif",
                    Ldown,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )
            if self.config.output_sh:
                common.save_raster(
                    self.config.output_dir + "/Shadow_" + time_code + ".tif",
                    shadow,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )
            if self.config.output_kdiff:
                common.save_raster(
                    self.config.output_dir + "/Kdiff_" + time_code + ".tif",
                    dRad,
                    dsm_trf_arr,
                    dsm_crs_wkt,
                    dsm_nd_val,
                )

            # Sky view image of patches
            if (self.config.aniso_path) and (i == 0) and (self.poi_pixel_xys is not None):
                for k in range(self.poi_pixel_xys.shape[0]):
                    Lsky_patch_characteristics[:, 2] = patch_characteristics[:, k]
                    skyviewimage_out = self.config.output_dir + "/POI_" + str(self.poi_names[k]) + ".png"
                    PolarBarPlot(
                        Lsky_patch_characteristics,
                        environ_data.altitude[i],
                        environ_data.azimuth[i],
                        "Hemisphere partitioning",
                        skyviewimage_out,
                        0,
                        5,
                        0,
                    )

        # Save POI results
        if self.poi_results:
            self.save_poi_results(dsm_trf_arr, dsm_crs_wkt)

        # Save WOI results
        if self.woi_results:
            self.save_woi_results(dsm_trf_arr, dsm_crs_wkt)

        # Save Tree Planter results
        if self.config.output_tree_planter:
            pos = 1 if self.params.Tmrt_params.Value.posture == "Standing" else 0

            settingsHeader = [
                "UTC",
                "posture",
                "onlyglobal",
                "landcover",
                "anisotropic",
                "cylinder",
                "albedo_walls",
                "albedo_ground",
                "emissivity_walls",
                "emissivity_ground",
                "absK",
                "absL",
                "elevation",
                "patch_option",
            ]
            settingsFmt = (
                "%i",
                "%i",
                "%i",
                "%i",
                "%i",
                "%i",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%1.2f",
                "%i",
            )
            settingsData = np.array(
                [
                    [
                        int(self.config.utc),
                        pos,
                        self.config.only_global,
                        self.config.use_landcover,
                        self.config.use_aniso,
                        self.config.person_cylinder,
                        self.params.Albedo.Effective.Value.Walls,
                        self.params.Albedo.Effective.Value.Cobble_stone_2014a,
                        self.params.Emissivity.Value.Walls,
                        self.params.Emissivity.Value.Cobble_stone_2014a,
                        self.params.Tmrt_params.Value.absK,
                        self.params.Tmrt_params.Value.absL,
                        self.location["altitude"],
                        shadow_mats.patch_option,
                    ]
                ]
            )
            np.savetxt(
                self.config.output_dir + "/treeplantersettings.txt",
                settingsData,
                fmt=settingsFmt,
                header=", ".join(settingsHeader),
                delimiter=" ",
            )

        # Save average Tmrt raster
        tmrtplot = tmrtplot / self.iters_total
        common.save_raster(
            self.config.output_dir + "/Tmrt_average.tif",
            tmrtplot,
            dsm_trf_arr,
            dsm_crs_wkt,
            dsm_nd_val,
        )



def Lvikt_veg(svf,svfveg,svfaveg,vikttot):

    # Least
    viktonlywall=(vikttot-(63.227*svf**6-161.51*svf**5+156.91*svf**4-70.424*svf**3+16.773*svf**2-0.4863*svf))/vikttot

    viktaveg=(vikttot-(63.227*svfaveg**6-161.51*svfaveg**5+156.91*svfaveg**4-70.424*svfaveg**3+16.773*svfaveg**2-0.4863*svfaveg))/vikttot

    viktwall=viktonlywall-viktaveg

    svfvegbu=(svfveg+svf-1)  # Vegetation plus buildings
    viktsky=(63.227*svfvegbu**6-161.51*svfvegbu**5+156.91*svfvegbu**4-70.424*svfvegbu**3+16.773*svfvegbu**2-0.4863*svfvegbu)/vikttot
    viktrefl=(vikttot-(63.227*svfvegbu**6-161.51*svfvegbu**5+156.91*svfvegbu**4-70.424*svfvegbu**3+16.773*svfvegbu**2-0.4863*svfvegbu))/vikttot
    viktveg=(vikttot-(63.227*svfvegbu**6-161.51*svfvegbu**5+156.91*svfvegbu**4-70.424*svfvegbu**3+16.773*svfvegbu**2-0.4863*svfvegbu))/vikttot
    viktveg=viktveg-viktwall

    return viktveg,viktwall,viktsky,viktrefl


# %%
from importlib import reload
from pathlib import Path

import geopandas as gpd
import pandas as pd

from umep import (
    common,
    shadow_generator_algorithm,
    skyviewfactor_algorithm,
    solweig_algorithm,
    wall_heightaspect_algorithm,
)

reload(solweig_algorithm)

#
bbox = [476070, 4203550, 477110, 4204330]
working_folder = "temp/tests/athens"
pixel_resolution = 1  # metres
working_crs = 2100

working_path = Path(working_folder).absolute()
working_path.mkdir(parents=True, exist_ok=True)
working_path_str = str(working_path)

# input files for computing
bldgs_path = "tests/data/athens/bldgs_clip.gpkg"
trees_path = "tests/data/athens/trees_clip.gpkg"
tree_canopies_path = "tests/data/athens/tree_canopies_clip.gpkg"

# %%
# create DSM if necessary
# the algorithm creates a new folder
# if this folder / file already exists then this step is skipped
# this saves unnecessary repetition
# if wanting to repeat, then delete (or rename) the folders / files
# same idea for remaining cells
if not Path.exists(working_path / "DSM.tif") or not Path.exists(working_path / "CDSM.tif"):
    # buildings
    bldgs_gdf = gpd.read_file(bldgs_path)
    bldgs_gdf = bldgs_gdf.to_crs(working_crs)
    # convert to raster by "burning" height column
    bldgs_rast, bldgs_transf = common.rasterise_gdf(
        bldgs_gdf,
        "geometry",
        "Height",
        bbox=bbox,
        pixel_size=pixel_resolution,
    )
    # save to geotiff
    common.save_raster(working_path_str + "/DSM.tif", bldgs_rast, bldgs_transf.to_gdal(), bldgs_gdf.crs.to_wkt())
    # GDF1 trees - load and buffer by diameter
    trees_gdf = gpd.read_file(trees_path)
    trees_gdf = trees_gdf.to_crs(working_crs)
    trees_gdf["geometry"] = trees_gdf["geometry"].buffer(trees_gdf["dia"])
    # GDF2 canopies - load and assign height column "ht" to match bldg ht column name
    tree_canopies_gdf = gpd.read_file(tree_canopies_path)
    tree_canopies_gdf = tree_canopies_gdf.to_crs(working_crs)
    tree_canopies_gdf["ht"] = 6  # set ht
    green_gdf = pd.concat(
        [
            trees_gdf[["ht", "geometry"]],  # GDF1
            tree_canopies_gdf[["ht", "geometry"]],  # GDF2
        ],
        ignore_index=True,
    )
    # create new GDF
    green_gdf = gpd.GeoDataFrame(green_gdf, geometry="geometry", crs=trees_gdf.crs)
    # rasterise by burning ht column
    veg_rast, veg_transf = common.rasterise_gdf(
        green_gdf,
        "geometry",
        "ht",
        bbox=bbox,
        pixel_size=pixel_resolution,
    )
    # flatten canopy raster where overlapping buildings
    veg_rast[bldgs_rast > 0] = 0
    # save
    common.save_raster(working_path_str + "/CDSM.tif", veg_rast, veg_transf.to_gdal(), trees_gdf.crs.to_wkt())

# %%
# wall info for SOLWEIG
if not Path.exists(working_path / "walls"):
    wall_heightaspect_algorithm.generate_wall_hts(
        dsm_path=working_path_str + "/DSM.tif",
        bbox=bbox,
        out_dir=working_path_str + "/walls",
    )

# %%
# shadows (not required for SOLWEIG)
for shadow_date_Ymd in ["2024-03-21"]:
    shadow_dir_name = f"shadow_{shadow_date_Ymd}"
    if not Path.exists(working_path / shadow_dir_name):
        shadow_generator_algorithm.generate_shadows(
            dsm_path=working_path_str + "/DSM.tif",
            # target date to caculate shadows for
            shadow_date_Ymd=shadow_date_Ymd,  # %Y-%m-%d"
            wall_ht_path=working_path_str + "/walls/wall_hts.tif",
            wall_aspect_path=working_path_str + "/walls/wall_aspects.tif",
            bbox=bbox,
            out_dir=working_path_str + "/" + shadow_dir_name,
            # if wanting a specific time then specify below
            # otherwise computed for intervals
            shadow_time_HM=None,  # "%H:%M"
            time_interval_M=60,  # interval in minutes - if not computing specific time
            veg_dsm_path=working_path_str + "/CDSM.tif",
        )

# %%
# skyview factor for SOLWEIG
if not Path.exists(working_path / "svf"):
    skyviewfactor_algorithm.generate_svf(
        dsm_path=working_path_str + "/DSM.tif",
        bbox=bbox,
        out_dir=working_path_str + "/svf",
        cdsm_path=working_path_str + "/CDSM.tif",
        trans_veg=5,
    )

# %%
# POIs for sampling UTCI
pois_gdf = gpd.read_file("tests/data/athens/pois.gpkg")
# iter EPWs
for epw_path, solweig_dir_name, start_date_Ymd, end_date_Ymd in [
    (
        "tests/data/athens/athens_2050.epw",
        "solweig_2050_07",
        "2050-07-21",
        "2050-07-22",
    ),
]:
    if not Path.exists(working_path / solweig_dir_name):
        # run algorithm - requires paths to files calculated in previous steps
        solweig_algorithm.generate_solweig(
            dsm_path=working_path_str + "/DSM.tif",
            wall_ht_path=working_path_str + "/walls/wall_hts.tif",
            wall_aspect_path=working_path_str + "/walls/wall_aspects.tif",
            svf_path=working_path_str + "/svf/svfs.zip",
            # the EPW file
            epw_path=epw_path,
            # the bounding box
            bbox=bbox,
            # the output directory
            out_dir=working_path_str + "/" + solweig_dir_name,
            # start and end date for range to filter from EPW
            start_date_Ymd=start_date_Ymd,  # year must match EPW file!
            end_date_Ymd=end_date_Ymd,
            # hours to run
            hours=[7, 11, 15, 19],
            # canopy DSM
            veg_dsm_path=working_path_str + "/CDSM.tif",
            # optional POIs GDF
            pois_gdf=pois_gdf,
            trans_veg=5,
        )

# %%



# Generated with SMOP  0.41
from libsmop import *
# LupMeas_abs.m

    
@function
def LupMeas_abs(L=None,D=None,Lup=None,*args,**kwargs):
    varargin = LupMeas_abs.varargin
    nargin = LupMeas_abs.nargin

    #Used to calculate the estimated outgoing terrestrial longwave absorbed by the
#cylinder  for use in the Rabs (Estimation) Model with inputs Ta (air temperature),
#L (length of cylinder, m), D (diameter of cylinder, cm)
    
    Acyl=CRT_Acyl(L,D)
# LupMeas_abs.m:7
    
    epsilon=0.95
# LupMeas_abs.m:9
    
    
    LupMeas_abs=multiply(multiply(multiply(epsilon,Lup),0.5),Acyl)
# LupMeas_abs.m:11
    #Written August, 2006 by Natasha Kenny
#Updated/Commented Dec 2015 Jenni Vanos


import numpy as np
from qgis.core import QgsVectorLayer
from osgeo import gdal, osr
from osgeo.gdalconst import *

def pointOfInterest(poilyr, poi_field, scale, gdal_dsm):

    (dsm_minx, dsm_x_size, dsm_x_rotation, dsm_miny, dsm_y_rotation, dsm_y_size) = gdal_dsm.GetGeoTransform() #TODO: fix for standalone

    # header = 'yyyy id   it imin dectime altitude azimuth    Ta' 
    poilyr = QgsVectorLayer(poilyr, 'point', 'ogr')
    idx = poilyr.fields().indexFromName(poi_field)
    numfeat = poilyr.featureCount()
    poiname = []
    poisxy = np.zeros((numfeat, 3)) - 999
    ind = 0
    for f in poilyr.getFeatures():  # looping through each POI
        y = f.geometry().centroid().asPoint().y()
        x = f.geometry().centroid().asPoint().x()

        poiname.append(f.attributes()[idx])
        poisxy[ind, 0] = ind
        # poisxy[ind, 1] = np.round((x - minx) * scale)

        xcoordinate = np.floor((x - dsm_minx) * scale)
        ycoordinate = np.floor((dsm_miny - y) * scale)

        # if miny >= 0:
        #     poisxy[ind, 2] = np.round((miny + rows * (1. / scale) - y) * scale)
        # else:
        #     poisxy[ind, 2] = np.round((miny + rows * (1. / scale) - y) * scale)
            # print('y = ' + str(np.round((miny + rows * (1. / scale) - y) * scale)))

        poisxy[ind, 1] = xcoordinate
        poisxy[ind, 2] = ycoordinate

        ind += 1

    # for k in range(0, poisxy.shape[0]):
    #     poi_save = []  # np.zeros((1, 33))
    #     data_out = outputDir + '/POI_' + str(poiname[k]) + '.txt'
    #     np.savetxt(data_out, poi_save,  delimiter=' ', header=header, comments='')

    return poisxy, poiname 

def fillWallOfInterest(voxelTable, voxelHeight, woisxy, woiname, outputDir, i, YYYY, jday, hours, minu, dectime, Ta, svf):

    if not woisxy is None:
        for k in range(0, woisxy.shape[0]):
            tempTable = voxelTable.loc[((voxelTable['ypos'] == woisxy[k, 2]) & (voxelTable['xpos'] == woisxy[k, 1]) & voxelTable['voxelHeight'] == voxelHeight)].copy()
            output_vars = np.array(['wallTemperature', 'K_in', 'L_in', 'wallTemperatureSolweigOld', 
                                    'Lwallsun', 'Lwallsh', 'Lrefl', 'Lveg', 'Lground', 'Lsky',
                                    'esky', 'F_sh', 'wallSun',
                                    'svfbu', 'svfveg', 'svfaveg'])
            
            counter = 0
            for temp_var in output_vars:
                temp_out = tempTable[temp_var].to_numpy()
                if counter == 0:
                    temp_all = temp_out.copy()
                else:
                    temp_all = np.concatenate([temp_all, temp_out])
                    
            wall_data = np.zeros((1, 7 + temp_all.shape[0]))
            # Part of file name (wallid), i.e. WOI_wallid.txt
            woiwallId = voxelTable.loc[((voxelTable['ypos'] == woisxy[k, 2]) & (voxelTable['xpos'] == woisxy[k, 1])), 'wallId'].to_numpy()[0]                    
            # print('wall id = ' + str(woiwallId) + ' and direction = ' + str(woiname[k]))
            data_out = outputDir + '/WOI_' + str(woiname[k]) + '_height' + str(voxelHeight) + '.txt'                    
            if i == 0:
                print('wall id = ' + str(woiwallId) + ' and direction = ' + str(woiname[k]))
                # Output file header
                header = 'yyyy id   it imin dectime Ta  SVF '
                voxelHeader = ''
                for temp_header in output_vars:
                    header += '    ' + temp_header
                #header = header + voxelHeader

                # Part of file name (wallid), i.e. WOI_wallid.txt
                # woiname = voxelTable.loc[((voxelTable['ypos'] == woisxy[k, 2]) & (voxelTable['xpos'] == woisxy[k, 1])), 'wallId'].to_numpy()[0]
                woi_save = []  # 
                # data_out = outputDir + '/WOI_' + str(woiname) + '.txt'
                np.savetxt(data_out, woi_save,  delimiter=' ', header=header, comments='')                        
            # Fill wall_data with variables
            wall_data[0, 0] = YYYY[0][i] 
            wall_data[0, 1] = jday[0][i]
            wall_data[0, 2] = hours[i]
            wall_data[0, 3] = minu[i]
            wall_data[0, 4] = dectime[i]
            wall_data[0, 5] = Ta[i]
            wall_data[0, 6] = svf[int(woisxy[k, 2]), int(woisxy[k, 1])]
            wall_data[0, 7:] = temp_all
            # wall_data[0, 7:] = temp_wall

            # Num format for output file data
            # woi_numformat = '%d %d %d %d %.5f %.2f %.2f' + ' %.2f' * temp_wall.shape[0]
            woi_numformat = '%d %d %d %d %.5f %.2f %.2f' + ' %.2f' * temp_all.shape[0]
            # Open file, add data, save
            f_handle = open(data_out, 'ab')
            np.savetxt(f_handle, wall_data, fmt=woi_numformat)
            f_handle.close()    


import datetime as dt
from builtins import range

import numpy as np
from tqdm import tqdm

from umep import common
from umep.util import shadowingfunctions as shadow
from umep.util.SEBESOLWEIGCommonFiles import sun_position as sp
from umep.util.SEBESOLWEIGCommonFiles.shadowingfunction_wallheight_13 import (
    shadowingfunction_wallheight_13,
)
from umep.util.SEBESOLWEIGCommonFiles.shadowingfunction_wallheight_23 import (
    shadowingfunction_wallheight_23,
)


def dailyshading(
    dsm,
    vegdsm,
    vegdsm2,
    scale,
    lon,
    lat,
    dsm_width,
    dsm_height,
    tv,
    UTC,
    usevegdem,
    timeInterval,
    onetime,
    folder,
    dsm_transf,
    dsm_crs,
    trans,
    dst,
    wallshadow,
    wheight,
    waspect,
):
    # lon = lonlat[0]
    # lat = lonlat[1]
    year = tv[0]
    month = tv[1]
    day = tv[2]

    alt = np.median(dsm)
    location = {"longitude": lon, "latitude": lat, "altitude": alt}
    if usevegdem == 1:
        psi = trans
        # amaxvalue
        vegmax = vegdsm.max()
        amaxvalue = dsm.max() - dsm.min()
        amaxvalue = np.maximum(amaxvalue, vegmax)

        # Elevation vegdsms if buildingDSM includes ground heights
        vegdem = vegdsm + dsm
        vegdem[vegdem == dsm] = 0
        vegdem2 = vegdsm2 + dsm
        vegdem2[vegdem2 == dsm] = 0

        # Bush separation
        bush = np.logical_not((vegdem2 * vegdem)) * vegdem

    shtot = np.zeros((dsm_height, dsm_width))

    if onetime == 1:
        itera = 1
    else:
        itera = int(1440 / timeInterval)

    alt = np.zeros(itera)
    azi = np.zeros(itera)
    hour = int(0)
    index = 0
    time = dict()
    time["UTC"] = UTC

    if wallshadow == 1:
        walls = wheight
        dirwalls = waspect
    else:
        walls = np.zeros((dsm_height, dsm_width))
        dirwalls = np.zeros((dsm_height, dsm_width))

    for i in tqdm(range(0, itera)):
        if onetime == 0:
            minu = int(timeInterval * i)
            if minu >= 60:
                hour = int(np.floor(minu / 60))
                minu = int(minu - hour * 60)
        else:
            minu = tv[4]
            hour = tv[3]

        doy = day_of_year(year, month, day)

        ut_time = (
            doy
            - 1.0
            + ((hour - dst) / 24.0)
            + (minu / (60.0 * 24.0))
            + (0.0 / (60.0 * 60.0 * 24.0))
        )

        if ut_time < 0:
            year = year - 1
            month = 12
            day = 31
            doy = day_of_year(year, month, day)
            ut_time = ut_time + doy - 1

        HHMMSS = dectime_to_timevec(ut_time)
        time["year"] = year
        time["month"] = month
        time["day"] = day
        time["hour"] = HHMMSS[0]
        time["min"] = HHMMSS[1]
        time["sec"] = HHMMSS[2]

        sun = sp.sun_position(time, location)
        alt[i] = 90.0 - sun["zenith"]
        azi[i] = sun["azimuth"]

        if time["sec"] == 59:  # issue 228 and 256
            time["sec"] = 0
            time["min"] = time["min"] + 1
            if time["min"] == 60:
                time["min"] = 0
                time["hour"] = time["hour"] + 1
                if time["hour"] == 24:
                    time["hour"] = 0

        time_vector = dt.datetime(
            year, month, day, time["hour"], time["min"], time["sec"]
        )
        timestr = time_vector.strftime("%Y%m%d_%H%M")
        if alt[i] > 0:
            if wallshadow == 1:  # Include wall shadows (Issue #121)
                if usevegdem == 1:
                    vegsh, sh, _, wallsh, _, wallshve, _, _ = (
                        shadowingfunction_wallheight_23(
                            dsm,
                            vegdem,
                            vegdem2,
                            azi[i],
                            alt[i],
                            scale,
                            amaxvalue,
                            bush,
                            walls,
                            dirwalls * np.pi / 180.0,
                        )
                    )
                    # create output folders
                    sh = sh - (1 - vegsh) * (1 - psi)
                    if onetime == 0:
                        filenamewallshve = (
                            folder
                            + "/facade_shdw_veg/facade_shdw_veg_"
                            + timestr
                            + "_LST.tif"
                        )
                        common.save_raster(
                            filenamewallshve, wallshve, dsm_transf, dsm_crs
                        )
                else:
                    sh, wallsh, _, _, _ = shadowingfunction_wallheight_13(
                        dsm, azi[i], alt[i], scale, walls, dirwalls * np.pi / 180.0
                    )
                    # shtot = shtot + sh

                if onetime == 0:
                    filename = (
                        folder + "/shadow_ground/shadow_ground_" + timestr + "_LST.tif"
                    )
                    common.save_raster(filename, sh, dsm_transf, dsm_crs)
                    filenamewallsh = (
                        folder
                        + "/facade_shdw_bldgs/facade_shdw_bldgs_"
                        + timestr
                        + "_LST.tif"
                    )
                    common.save_raster(filenamewallsh, wallsh, dsm_transf, dsm_crs)

            else:
                if usevegdem == 0:
                    sh = shadow.shadowingfunctionglobalradiation(
                        dsm, azi[i], alt[i], scale, 0
                    )
                    # shtot = shtot + sh
                else:
                    shadowresult = shadow.shadowingfunction_20(
                        dsm,
                        vegdem,
                        vegdem2,
                        azi[i],
                        alt[i],
                        scale,
                        amaxvalue,
                        bush,
                        0,
                    )
                    vegsh = shadowresult["vegsh"]
                    sh = shadowresult["sh"]
                    sh = sh - (1 - vegsh) * (1 - psi)
                    # vegshtot = vegshtot + sh

                if onetime == 0:
                    filename = folder + "/Shadow_" + timestr + "_LST.tif"
                    common.save_raster(filename, sh, dsm_transf, dsm_crs)

            shtot = shtot + sh
            index += 1

    shfinal = shtot / index

    if wallshadow == 1:
        if onetime == 1:
            filenamewallsh = (
                folder + "/facade_shdw_bldgs/facade_shdw_bldgs_" + timestr + "_LST.tif"
            )
            common.save_raster(filenamewallsh, wallsh, dsm_transf, dsm_crs)
            if usevegdem == 1:
                filenamewallshve = (
                    folder + "/facade_shdw_veg/facade_shdw_veg_" + timestr + "_LST.tif"
                )
                common.save_raster(filenamewallshve, wallshve, dsm_transf, dsm_crs)

    shadowresult = {"shfinal": shfinal, "time_vector": time_vector}

    return shadowresult


def day_of_year(yy, month, day):
    if (yy % 4) == 0:
        if (yy % 100) == 0:
            if (yy % 400) == 0:
                leapyear = 1
            else:
                leapyear = 0
        else:
            leapyear = 1
    else:
        leapyear = 0

    if leapyear == 1:
        dayspermonth = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    else:
        dayspermonth = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

    doy = np.sum(dayspermonth[0 : month - 1]) + day

    return doy


def dectime_to_timevec(dectime):
    # This subroutine converts dectime to individual hours, minutes and seconds

    doy = np.floor(dectime)

    DH = dectime - doy
    HOURS = int(24 * DH)

    DM = 24 * DH - HOURS
    MINS = int(60 * DM)

    DS = 60 * DM - MINS
    SECS = int(60 * DS)

    return (HOURS, MINS, SECS)



import os
from configparser import ConfigParser

def read_solweig_config(config_file_path):
    if not os.path.exists(config_file_path):
        raise ValueError("The config file '{}' could not be found.".format(config_file_path))

    config = ConfigParser()
    config.read(config_file_path)
    return config.defaults()


'''
input: 
TargetDict: dictonary with all relevant inputs
refdir = outputfolder      
'''

def write_solweig_config(configDict, refdir):
    f = open(refdir, "w")

    f.write("\n")
    f.write("[DEFAULT]\n")
    f.write("\n")
    f.write("### INPUTS ###\n")
    f.write("# output path\n")
    f.write("output_dir={}\n".format(configDict['output_dir']))
    f.write("# working path\n")
    f.write("working_dir={}\n".format(configDict['working_dir']))
    f.write("# parameters json file\n")
    f.write("para_json_path={}\n".format(configDict['para_json_path']))
    f.write("# Input ground and building dsm\n")
    f.write("filepath_dsm={}\n".format(configDict['filepath_dsm']))
    f.write("# Input vegetation dsm\n")
    f.write("filepath_cdsm={}\n".format(configDict['filepath_cdsm']))
    f.write("# Input trunkzone vegetation dsm\n")
    f.write("filepath_tdsm={}\n".format(configDict['filepath_tdsm']))    
    f.write("# Input Digital Elevation Model\n")
    f.write("filepath_dem={}\n".format(configDict['filepath_dem'])) 
    f.write("# Input Land cover dataset\n")
    f.write("filepath_lc={}\n".format(configDict['filepath_lc'])) 
    f.write("# Input wall height raster\n")
    f.write("filepath_wh={}\n".format(configDict['filepath_wh'])) 
    f.write("# Input wall aspect raster\n")
    f.write("filepath_wa={}\n".format(configDict['filepath_wa']))
    f.write("# Skyview factor files\n")
    f.write("input_svf={}\n".format(configDict['input_svf']))
    f.write("# Input file for anisotrophic sky\n")
    f.write("input_aniso={}\n".format(configDict['input_aniso']))
    f.write("# Point of Interest file for ground\n")
    f.write("poi_file={}\n".format(configDict['poi_file']))
    f.write("poi_field={}\n".format(configDict['poi_field']))  
    f.write("# Input file for wall temperture scheme (Wallenberg et al. 2025)\n")
    f.write("input_wall={}\n".format(configDict['input_wall']))     
    f.write("# Point of Interest file for walls\n")
    f.write("woi_file={}\n".format(configDict['woi_file']))
    f.write("woi_field={}\n".format(configDict['woi_field']))   
    f.write("# input meteorolgical file (i.e. forcing file)\n")
    f.write("input_met={}\n".format(configDict['input_met']))            
    f.write("\n")
    f.write("## input settings ##\n")
    f.write("# dates\n")
    f.write("# option to execute solweig outside of osgeo/qgis environment\n")
    f.write("standalone={}\n".format(configDict['standalone'])) 
    f.write("# estimate diffuse and global shortwve radiation from global radiation (1)\n")
    f.write("onlyglobal={}\n".format(configDict['onlyglobal'])) 
    f.write("# use vegetation scheme (lindberg and grimmond, 2011, TAAC)\n")
    f.write("usevegdem={}\n".format(configDict['usevegdem'])) 
    f.write("# consider leaf on trees full year (1)\n")
    f.write("conifer_bool={}\n".format(configDict['conifer_bool'])) 
    f.write("# consider person as a cylinder (1) or a box (0)\n")
    f.write("cyl={}\n".format(configDict['cyl'])) 
    # f.write("# Location of model domain\n")
    # f.write("lat={}\n".format(configDict['lat'])) 
    # f.write("lon={}\n".format(configDict['lon'])) 
    f.write("#utc time given in meteorological forcing file\n")
    f.write("utc={}\n".format(configDict['utc'])) 
    # f.write("# scale between horisontal and vertical resolution (1 = 1 meter pixel reolution, 0.5 = 2)\n")
    # f.write("scale={}\n".format(configDict['scale'])) 
    # f.write("# DSM CRS coordinates of lower left corner\n")
    # f.write("minx={}\n".format(configDict['minx'])) 
    # f.write("miny={}\n".format(configDict['miny']))
    f.write("# Set to 1 if an EPW file is used as meteorological forcing data\n")
    f.write("useepwfile={}\n".format(configDict['useepwfile']))
    f.write("# land cover scheme activated (1) (Lindberg et al. 2016 UC)\n")
    f.write("landcover={}\n".format(configDict['landcover']))
    f.write("# use dem and dsm to identify building pixels (0) else, land cover will be used (0)\n")
    f.write("demforbuild={}\n".format(configDict['demforbuild']))    
    f.write("# use anisotropic sky (Wallenberg et al. 2020 and Wallenberg et al. 2023)\n")
    f.write("aniso={}\n".format(configDict['aniso']))
    f.write("# use wall surface temperature scheme (Wallenberg et al. 2025, GMD)\n")
    f.write("wallscheme={}\n".format(configDict['wallscheme']))
    f.write("# If building materials is not included in lc, then this is used for all buildings (Wood, Brick or Concrete)\n")
    f.write("walltype={}\n".format(configDict['walltype']))
    f.write("# output settings\n")
    f.write("outputtmrt={}\n".format(configDict['outputtmrt']))
    f.write("outputkup={}\n".format(configDict['outputkup']))
    f.write("outputkdown={}\n".format(configDict['outputkdown']))
    f.write("outputlup={}\n".format(configDict['outputlup']))
    f.write("outputldown={}\n".format(configDict['outputldown']))
    f.write("outputsh={}\n".format(configDict['outputsh']))    
    f.write("savebuild={}\n".format(configDict['savebuild']))   
    f.write("outputkdiff={}\n".format(configDict['outputkdiff']))       
    f.write("outputtreeplanter={}\n".format(configDict['outputtreeplanter']))   
    f.write("wallnetcdf={}\n".format(configDict['wallnetcdf']))   
    f.write("#-------------------------------------------------------\n")
    f.write("# dates - used if an EPW-file is used\n")    
    f.write("#-------------------------------------------------------\n")
    f.write("# year,month,day,hour	## the start date/time for period of interest\n")
    f.write("date1={}\n".format(configDict['date1']))
    f.write("# year,month,day,hour	# end date/time\n")
    f.write("date2={}\n".format(configDict['date2']))
    f.write("\n")

    f.close()



# Generated with SMOP  0.41
from libsmop import *
# CRT_Acs.m

    
@function
def CRT_Acs(L=None,D=None,*args,**kwargs):
    varargin = CRT_Acs.varargin
    nargin = CRT_Acs.nargin

    # vertical cross sectional area of the cylinder in m
    
    # Written by Jenni Vanos, Aug 2009, to find the XC area of a cylinder,
# where L = length (m) and D = diameter (m)- #for CRT L = ~0.10m and D =
# 0.01m - ratio is most important part to keep constant for changing
# cylidner size. 
# 
#Used in Kb_abs = (1-alpha).* (Kb .* sind(zen)).* Acs;
    
    Acs=multiply(D,L)
# CRT_Acs.m:13



def Kvikt_veg(svf,svfveg,vikttot):

    # Least
    viktwall=(vikttot-(63.227*svf**6-161.51*svf**5+156.91*svf**4-70.424*svf**3+16.773*svf**2-0.4863*svf))/vikttot
    
    svfvegbu=(svfveg+svf-1)  # Vegetation plus buildings
    viktveg=(vikttot-(63.227*svfvegbu**6-161.51*svfvegbu**5+156.91*svfvegbu**4-70.424*svfvegbu**3+16.773*svfvegbu**2-0.4863*svfvegbu))/vikttot
    viktveg=viktveg-viktwall
    
    return viktveg,viktwall


"""
/***************************************************************************
 ProcessingUMEP
                                 A QGIS plugin
 UMEP for processing toolbox
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2020-04-02
        copyright            : (C) 2020 by Fredrik Lindberg
        email                : fredrikl@gvc.gu.se
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

 "This algorithm identiies wall pixels and "
"their height from ground and building digital surface models (DSM) by using a filter as "
"presented by Lindberg et al. (2015a). Optionally, wall aspect can also be estimated using "
"a specific linear filter as presented by Goodwin et al. (1999) and further developed by "
"Lindberg et al. (2015b) to obtain the wall aspect. Wall aspect is given in degrees where "
"a north facing wall pixel has a value of zero. The output of this plugin is used in other "
"UMEP plugins such as SEBE (Solar Energy on Building Envelopes) and SOLWEIG (SOlar LongWave "
"Environmental Irradiance Geometry model).\n"
"------------------ \n"
"Goodwin NR, Coops NC, Tooke TR, Christen A, Voogt JA (2009) Characterizing urban surface cover and structure with airborne lidar technology. Can J Remote Sens 35:297–309\n"
"Lindberg F., Grimmond, C.S.B. and Martilli, A. (2015a) Sunlit fractions on urban facets - Impact of spatial resolution and approach Urban Climate DOI: 10.1016/j.uclim.2014.11.006\n"
"Lindberg F., Jonsson, P. & Honjo, T. and Wästberg, D. (2015b) Solar energy on building envelopes - 3D modelling in a 2D environment Solar Energy 115 369–378"
"-------------\n"
"Full manual available via the <b>Help</b>-button."
"https://umep-docs.readthedocs.io/en/latest/pre-processor/Urban%20Geometry%20Wall%20Height%20and%20Aspect.html"
"""

__author__ = "Fredrik Lindberg"
__date__ = "2020-04-02"
__copyright__ = "(C) 2020 by Fredrik Lindberg"

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = "$Format:%H$"


from pathlib import Path

from umep import common
from umep.functions import wallalgorithms as wa


def generate_wall_hts(
    dsm_path: str,
    bbox: list[int, int, int, int],
    out_dir: str,
    wall_limit: float = 1,
):
    """ """
    dsm_rast, dsm_transf, dsm_crs, _dsm_nd = common.load_raster(dsm_path, bbox)
    dsm_scale = 1 / dsm_transf[1]

    out_path = Path(out_dir)
    out_path.mkdir(parents=True, exist_ok=True)
    out_path_str = str(out_path)

    walls = wa.findwalls(dsm_rast, wall_limit)
    common.save_raster(out_path_str + "/" + "wall_hts.tif", walls, dsm_transf, dsm_crs)

    dirwalls = wa.filter1Goodwin_as_aspect_v3(walls, dsm_scale, dsm_rast)
    common.save_raster(out_path_str + "/" + "wall_aspects.tif", dirwalls, dsm_transf, dsm_crs)



import numpy as np


def TsWaveDelay_2015a(gvfLup, firstdaytime, timeadd, timestepdec, Tgmap1):

    Tgmap0 = gvfLup  # current timestep
    if firstdaytime == 1:  # "first in morning"
        Tgmap1 = Tgmap0

    if timeadd >= (59/1440):  # more or equal to 59 min
        weight1 = np.exp(-33.27 * timeadd)  # surface temperature delay function - 1 step
        Tgmap1 = Tgmap0 * (1 - weight1) + Tgmap1 * weight1
        Lup = Tgmap1
        if timestepdec > (59/1440):
            timeadd = timestepdec
        else:
            timeadd = 0
    else:
        timeadd = timeadd + timestepdec
        weight1 = np.exp(-33.27 * timeadd)  # surface temperature delay function - 1 step
        Lup = (Tgmap0 * (1 - weight1) + Tgmap1 * weight1)

    return Lup, timeadd, Tgmap1


from __future__ import absolute_import
# from importdata import importdata
from . import sun_position as sp
#import sun_position as sp
import numpy as np
import datetime
import calendar

def Solweig_2015a_metdata_noload(inputdata, location, UTC):
    """
    This function is used to process the input meteorological file.
    It also calculates Sun position based on the time specified in the met-file

    :param inputdata:
    :param location:
    :param UTC:
    :return:
    """

    met = inputdata
    data_len = len(met[:, 0])
    dectime = met[:, 1]+met[:, 2] / 24 + met[:, 3] / (60*24.)
    dectimemin = met[:, 3] / (60*24.)
    if data_len == 1:
        halftimestepdec = 0
    else:
        halftimestepdec = (dectime[1] - dectime[0]) / 2.
    time = dict()
    time['sec'] = 0
    time['UTC'] = UTC
    sunmaximum = 0.
    leafon1 = 97  #TODO this should change
    leafoff1 = 300  #TODO this should change

    # initialize matrices
    altitude = np.empty(shape=(1, data_len))
    azimuth = np.empty(shape=(1, data_len))
    zen = np.empty(shape=(1, data_len))
    jday = np.empty(shape=(1, data_len))
    YYYY = np.empty(shape=(1, data_len))
    leafon = np.empty(shape=(1, data_len))
    altmax = np.empty(shape=(1, data_len))

    sunmax = dict()

    for i, row in enumerate(met[:, 0]):
        if met[i, 1] == 221:
            test = 4
        YMD = datetime.datetime(int(met[i, 0]), 1, 1) + datetime.timedelta(int(met[i, 1]) - 1)
        # Finding maximum altitude in 15 min intervals (20141027)
        if (i == 0) or (np.mod(dectime[i], np.floor(dectime[i])) == 0):
            fifteen = 0.
            sunmaximum = -90.
            sunmax['zenith'] = 90.
            while sunmaximum <= 90. - sunmax['zenith']:
                sunmaximum = 90. - sunmax['zenith']
                fifteen = fifteen + 15. / 1440.
                HM = datetime.timedelta(days=(60*10)/1440.0 + fifteen)
                YMDHM = YMD + HM
                time['year'] = YMDHM.year
                time['month'] = YMDHM.month
                time['day'] = YMDHM.day
                time['hour'] = YMDHM.hour
                time['min'] = YMDHM.minute
                sunmax = sp.sun_position(time,location)
        altmax[0, i] = sunmaximum

        half = datetime.timedelta(days=halftimestepdec)
        H = datetime.timedelta(hours=met[i, 2])
        M = datetime.timedelta(minutes=met[i, 3])
        YMDHM = YMD + H + M - half
        time['year'] = YMDHM.year
        time['month'] = YMDHM.month
        time['day'] = YMDHM.day
        time['hour'] = YMDHM.hour
        time['min'] = YMDHM.minute
        sun = sp.sun_position(time, location)
        if (sun['zenith'] > 89.0) & (sun['zenith'] <= 90.0):    # Hopefully fixes weird values in Perez et al. when altitude < 1.0, i.e. close to sunrise/sunset
            sun['zenith'] = 89.0
        altitude[0, i] = 90. - sun['zenith']
        zen[0, i] = sun['zenith'] * (np.pi/180.)
        azimuth[0, i] = sun['azimuth']

        # day of year and check for leap year
        if calendar.isleap(time['year']):
            dayspermonth = np.atleast_2d([31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31])
        else:
            dayspermonth = np.atleast_2d([31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31])
        # jday[0, i] = np.sum(dayspermonth[0, 0:time['month']-1]) + time['day'] # bug when a new day 20191015
        YYYY[0, i] = met[i, 0]
        doy = YMD.timetuple().tm_yday
        jday[0, i] = doy
        if (doy > leafon1) | (doy < leafoff1):
            leafon[0, i] = 1
        else:
            leafon[0, i] = 0

    return YYYY, altitude, azimuth, zen, jday, leafon, dectime, altmax







import sys
from builtins import str

# This file prints out run information used for each specific run
from time import strftime

import qgis.core
from osgeo import gdal, osr


def writeRunInfo(
    folderPath,
    filepath_dsm,
    gdal_dsm,
    usevegdem,
    filePath_cdsm,
    trunkfile,
    filePath_tdsm,
    lat,
    lon,
    UTC,
    landcover,
    filePath_lc,
    metfileexist,
    filePath_metfile,
    metdata,
    plugin_dir,
    absK,
    absL,
    albedo_b,
    albedo_g,
    ewall,
    eground,
    onlyglobal,
    trunkratio,
    trans,
    rows,
    cols,
    pos,
    elvis,
    cyl,
    demforbuild,
    ani,
    wallScheme,
    effusivity,
):
    # with open(folderPath + '/RunInfoSOLWEIG.txt', 'w') as file:           	#FO#
    # FO#
    if metdata[0, 2] < 10:
        XH = "0"
    else:
        XH = ""
    if metdata[0, 3] < 10:
        XM = "0"
    else:
        XM = ""
    with open(
        folderPath
        + "/RunInfoSOLWEIG_"
        + str(int(metdata[0, 0]))
        + "_"
        + str(int(metdata[0, 1]))
        + "_"
        + XH
        + str(int(metdata[0, 2]))
        + XM
        + str(int(metdata[0, 3]))
        + ".txt",
        "w",
    ) as file:
        # FO#
        file.write(
            "This file provides run settings for the SOLWEIG run initiated at: " + strftime("%a, %d %b %Y %H:%M:%S")
        )
        file.write("\n")
        file.write("Version: " + "SOLWEIG v2022a")
        file.write("\n")
        file.write("Pyton version: " + str(sys.version))
        file.write("\n")
        file.write("GDAL version: " + str(gdal.__version__))
        file.write("\n")
        file.write("QGIS version: " + str(qgis.core.Qgis.QGIS_VERSION))
        file.write("\n")
        file.write("\n")
        file.write("SURFACE DATA")
        file.write("\n")
        file.write("Digital surface model (DSM): " + filepath_dsm)
        file.write("\n")
        file.write("Model domain: rows = " + str(rows) + ", columns = " + str(cols))
        file.write("\n")
        # get CRS
        prj = gdal_dsm.GetProjection()
        srs = osr.SpatialReference(wkt=prj)
        if srs.IsProjected:
            file.write("Projected reference system: " + srs.GetAttrValue("projcs"))
        file.write("\n")
        file.write("Geographical coordinate system: " + srs.GetAttrValue("geogcs"))
        file.write("\n")
        file.write("Latitude: " + str(lat))
        file.write("\n")
        file.write("Longitude: " + str(lon))
        file.write("\n")
        file.write("UTC: " + str(UTC))
        file.write("\n")
        if usevegdem == 1:
            file.write("Transmissivity of light through vegetation: " + str(trans))
            file.write("\n")
            file.write("Digital vegetation canopy model (CDSM): " + filePath_cdsm)
            file.write("\n")
            if trunkfile == 1:
                file.write("Digital vegetation trunk zone model (TDSM): " + filePath_tdsm)  # FO# zrunk -> trunk
                file.write("\n")
            else:
                file.write("Trunkzone estimated from CDSM")
                file.write("\n")
                file.write("Trunkzone as percent of canopy height: " + str(trunkratio))
                file.write("\n")
        else:
            file.write("Vegetation scheme inactive")
            file.write("\n")
        if landcover == 1:
            file.write("Landcover scheme active. Parameters taken from: " + plugin_dir + "/landcoverclasses_2016a.txt")
            file.write("\n")
            file.write("Landcover grid: " + filePath_lc)
            file.write("\n")
        else:
            file.write("Landcover scheme inactive")
            file.write("\n")
        file.write("\n")
        if demforbuild == 1:
            file.write("DEM used to identify buildings")
            file.write("\n")
        else:
            file.write("Land cover used to identify buildings")
            file.write("\n")
        file.write("\n")
        file.write("METEOROLOGICAL FORCING DATA")
        file.write("\n")
        if metfileexist == 1:
            file.write("Meteorological file: " + filePath_metfile)
            file.write("\n")
            if onlyglobal == 1:
                file.write("Diffuse and direct shortwave radiation estimated from global radiation")
                file.write("\n")
        else:
            file.write("Meteorological file not used")
            file.write("\n")  # FO# ' ' -> file.write('\n')
            file.write("Year: " + str(metdata[0, 0]))
            file.write("\n")
            file.write("Day of Year: " + str(metdata[0, 1]))
            file.write("\n")
            file.write("Hour: " + str(metdata[0, 2]))
            file.write("\n")
            file.write("Minute: " + str(metdata[0, 3]))
            file.write("\n")
            file.write("Air temperature: " + str(metdata[0, 11]))  # FO# Ait -> Air
            file.write("\n")
            file.write("Relative humidity: " + str(metdata[0, 10]))
            file.write("\n")
            file.write("Global radiation: " + str(metdata[0, 14]))
            file.write("\n")
            file.write("Diffuse radiation: " + str(metdata[0, 21]))
            file.write("\n")
            file.write("Direct radiation: " + str(metdata[0, 22]))
            file.write("\n")
        file.write("\n")
        file.write("HUMAN EXPOSURE PARAMETERS")
        file.write("\n")
        file.write("Absorption, shortwave radiation: " + str(absK))
        file.write("\n")
        file.write("Absorption, longwave radiation: " + str(absL))
        file.write("\n")
        if pos == 0:
            file.write("Posture of human body: Standing")
        else:
            file.write("Posture of human body: Sitting")
        file.write("\n")
        file.write("ENVIRONMENTAL PARAMETERS")
        file.write("\n")
        file.write("Albedo of walls: " + str(albedo_b))
        file.write("\n")
        file.write("Albedo of ground (not used if land cover scheme is active): " + str(albedo_g))
        file.write("\n")
        file.write("Emissivity (walls): " + str(ewall))
        file.write("\n")
        file.write("Emissivity of ground (not used if land cover scheme is active): " + str(eground))
        file.write("\n")
        file.write("\n")
        file.write("ADDITIONAL SETTINGS")
        file.write("\n")
        if elvis == 1:
            file.write("Sky emissivity adjusted according to Jonsson et al. (2005)")
            file.write("\n")
        if cyl == 1:
            file.write("Human considered as a standing cylinder")  # FO# '' -> standing
        else:
            file.write("Human considered as a standing cube")
        file.write("\n")
        if ani == 1:
            file.write(
                "Anisotropic sky diffuse shortwave (Perez et al. 1993) and longwave (Martin & Berdahl, 1984) radiation"
            )
        else:
            file.write("Isotropic sky")
        file.write("\n")
        if wallScheme:
            if effusivity.size == 1:
                file.write(
                    "Wall surface parameterization scheme. Wall effusivity set to "
                    + str(effusivity)
                    + " J m-2 s-0.5 K-1."
                )
            else:
                for i in range(effusivity.size):
                    if i == 0:
                        file.write("There are " + str(effusivity.size) + " wall types:")
                        file.write("\n")
                    file.write(
                        "Wall surface parameterization scheme. Wall effusivity set to "
                        + str(effusivity[i])
                        + " J m-2 s-0.5 K-1."
                    )
                    file.write("\n")
        file.write("\n")
        file.close()



from __future__ import absolute_import

import numpy as np
from .daylen import daylen
from ...util.SEBESOLWEIGCommonFiles.clearnessindex_2013b import clearnessindex_2013b
from ...util.SEBESOLWEIGCommonFiles.diffusefraction import diffusefraction
from ...util.SEBESOLWEIGCommonFiles.shadowingfunction_wallheight_13 import shadowingfunction_wallheight_13
from ...util.SEBESOLWEIGCommonFiles.shadowingfunction_wallheight_23 import shadowingfunction_wallheight_23
from .gvf_2018a import gvf_2018a
from .cylindric_wedge import cylindric_wedge
from .TsWaveDelay_2015a import TsWaveDelay_2015a
from .Kup_veg_2015a import Kup_veg_2015a
# from .Lside_veg_v2015a import Lside_veg_v2015a
# from .Kside_veg_v2019a import Kside_veg_v2019a
from .Kside_veg_v2022a import Kside_veg_v2022a
from ...util.SEBESOLWEIGCommonFiles.Perez_v3 import Perez_v3
from ...util.SEBESOLWEIGCommonFiles.create_patches import create_patches

# Anisotropic longwave
from .Lcyl_v2022a import Lcyl_v2022a
from .Lside_veg_v2022a import Lside_veg_v2022a
from copy import deepcopy

def Solweig_2022a_calc(i, dsm, scale, rows, cols, svf, svfN, svfW, svfE, svfS, svfveg, svfNveg, svfEveg, svfSveg,
                       svfWveg, svfaveg, svfEaveg, svfSaveg, svfWaveg, svfNaveg, vegdem, vegdem2, albedo_b, absK, absL,
                       ewall, Fside, Fup, Fcyl, altitude, azimuth, zen, jday, usevegdem, onlyglobal, buildings, location, psi,
                       landcover, lc_grid, dectime, altmax, dirwalls, walls, cyl, elvis, Ta, RH, radG, radD, radI, P,
                       amaxvalue, bush, Twater, TgK, Tstart, alb_grid, emis_grid, TgK_wall, Tstart_wall, TmaxLST,
                       TmaxLST_wall, first, second, svfalfa, svfbuveg, firstdaytime, timeadd, timestepdec, Tgmap1, 
                       Tgmap1E, Tgmap1S, Tgmap1W, Tgmap1N, CI, TgOut1, diffsh, shmat, vegshmat, vbshvegshmat, anisotropic_sky, asvf, patch_option):

#def Solweig_2021a_calc(i, dsm, scale, rows, cols, svf, svfN, svfW, svfE, svfS, svfveg, svfNveg, svfEveg, svfSveg,
#                       svfWveg, svfaveg, svfEaveg, svfSaveg, svfWaveg, svfNaveg, vegdem, vegdem2, albedo_b, absK, absL,
#                       ewall, Fside, Fup, Fcyl, altitude, azimuth, zen, jday, usevegdem, onlyglobal, buildings, location, psi,
#                       landcover, lc_grid, dectime, altmax, dirwalls, walls, cyl, elvis, Ta, RH, radG, radD, radI, P,
#                       amaxvalue, bush, Twater, TgK, Tstart, alb_grid, emis_grid, TgK_wall, Tstart_wall, TmaxLST,
#                       TmaxLST_wall, first, second, svfalfa, svfbuveg, firstdaytime, timeadd, timestepdec, Tgmap1, 
#                       Tgmap1E, Tgmap1S, Tgmap1W, Tgmap1N, CI, TgOut1, diffsh, ani):

    # This is the core function of the SOLWEIG model
    # 2016-Aug-28
    # Fredrik Lindberg, fredrikl@gvc.gu.se
    # Goteborg Urban Climate Group
    # Gothenburg University
    #
    # Input variables:
    # dsm = digital surface model
    # scale = height to pixel size (2m pixel gives scale = 0.5)
    # svf,svfN,svfW,svfE,svfS = SVFs for building and ground
    # svfveg,svfNveg,svfEveg,svfSveg,svfWveg = Veg SVFs blocking sky
    # svfaveg,svfEaveg,svfSaveg,svfWaveg,svfNaveg = Veg SVFs blocking buildings
    # vegdem = Vegetation canopy DSM
    # vegdem2 = Vegetation trunk zone DSM
    # albedo_b = building wall albedo
    # absK = human absorption coefficient for shortwave radiation
    # absL = human absorption coefficient for longwave radiation
    # ewall = Emissivity of building walls
    # Fside = The angular factors between a person and the surrounding surfaces
    # Fup = The angular factors between a person and the surrounding surfaces
    # Fcyl = The angular factors between a culidric person and the surrounding surfaces
    # altitude = Sun altitude (degree)
    # azimuth = Sun azimuth (degree)
    # zen = Sun zenith angle (radians)
    # jday = day of year
    # usevegdem = use vegetation scheme
    # onlyglobal = calculate dir and diff from global shortwave (Reindl et al. 1990)
    # buildings = Boolena grid to identify building pixels
    # location = geographic location
    # height = height of measurements point (center of gravity of human)
    # psi = 1 - Transmissivity of shortwave through vegetation
    # landcover = use landcover scheme !!!NEW IN 2015a!!!
    # lc_grid = grid with landcoverclasses
    # lc_class = table with landcover properties
    # dectime = decimal time
    # altmax = maximum sun altitude
    # dirwalls = aspect of walls
    # walls = one pixel row outside building footprint. height of building walls
    # cyl = consider man as cylinder instead of cude
    # elvis = dummy
    # Ta = air temp
    # RH
    # radG = global radiation
    # radD = diffuse
    # radI = direct
    # P = pressure
    # amaxvalue = max height of buildings
    # bush = grid representing bushes
    # Twater = temperature of water (daily)
    # TgK, Tstart, TgK_wall, Tstart_wall, TmaxLST,TmaxLST_wall, 
    # alb_grid, emis_grid = albedo and emmissivity on ground
    # first, second = conneted to old Ts model (source area based on Smidt et al.)
    # svfalfa = SVF recalculated to angle
    # svfbuveg = complete SVF 
    # firstdaytime, timeadd, timestepdec, Tgmap1, Tgmap1E, Tgmap1S, Tgmap1W, Tgmap1N, 
    # CI = Clearness index
    # TgOut1 = old Ts model
    # diffsh, ani = Used in anisotrpic models (Wallenberg et al. 2019, 2022)

    # # # Core program start # # #
    # Instrument offset in degrees
    t = 0.

    # Stefan Bolzmans Constant
    SBC = 5.67051e-8

    # Find sunrise decimal hour - new from 2014a
    _, _, _, SNUP = daylen(jday, location['latitude'])

    # Vapor pressure
    ea = 6.107 * 10 ** ((7.5 * Ta) / (237.3 + Ta)) * (RH / 100.)

    # Determination of clear - sky emissivity from Prata (1996)
    msteg = 46.5 * (ea / (Ta + 273.15))
    esky = (1 - (1 + msteg) * np.exp(-((1.2 + 3.0 * msteg) ** 0.5))) + elvis  # -0.04 old error from Jonsson et al.2006

    if altitude > 0: # # # # # # DAYTIME # # # # # #
        # Clearness Index on Earth's surface after Crawford and Dunchon (1999) with a correction
        #  factor for low sun elevations after Lindberg et al.(2008)
        I0, CI, Kt, I0et, CIuncorr = clearnessindex_2013b(zen, jday, Ta, RH / 100., radG, location, P)
        if (CI > 1) or (CI == np.inf):
            CI = 1

        # Estimation of radD and radI if not measured after Reindl et al.(1990)
        if onlyglobal == 1:
            I0, CI, Kt, I0et, CIuncorr = clearnessindex_2013b(zen, jday, Ta, RH / 100., radG, location, P)
            if (CI > 1) or (CI == np.inf):
                CI = 1

            radI, radD = diffusefraction(radG, altitude, Kt, Ta, RH)

        # Diffuse Radiation
        # Anisotropic Diffuse Radiation after Perez et al. 1993
        if anisotropic_sky == 1:
            patchchoice = 1
            zenDeg = zen*(180/np.pi)
            # Relative luminance
            lv, pc_, pb_ = Perez_v3(zenDeg, azimuth, radD, radI, jday, patchchoice, patch_option)   
            # Total relative luminance from sky, i.e. from each patch, into each cell
            aniLum = np.zeros((rows, cols))
            for idx in range(lv.shape[0]):
                aniLum += diffsh[:,:,idx] * lv[idx,2]     

            dRad = aniLum * radD   # Total diffuse radiation from sky into each cell
        else:
            dRad = radD * svfbuveg
            patchchoice = 1
            # zenDeg = zen*(180/np.pi)
            lv = None
            # lv, pc_, pb_ = Perez_v3(zenDeg, azimuth, radD, radI, jday, patchchoice, patch_option)   # Relative luminance

        # Shadow  images
        if usevegdem == 1:
            vegsh, sh, _, wallsh, wallsun, wallshve, _, facesun = shadowingfunction_wallheight_23(dsm, vegdem, vegdem2,
                                        azimuth, altitude, scale, amaxvalue, bush, walls, dirwalls * np.pi / 180.)
            shadow = sh - (1 - vegsh) * (1 - psi)
        else:
            sh, wallsh, wallsun, facesh, facesun = shadowingfunction_wallheight_13(dsm, azimuth, altitude, scale,
                                                                                   walls, dirwalls * np.pi / 180.)
            shadow = sh

        # # # Surface temperature parameterisation during daytime # # # #
        # new using max sun alt.instead of  dfm
        # Tgamp = (TgK * altmax - Tstart) + Tstart # Old
        Tgamp = TgK * altmax + Tstart # Fixed 2021
        # Tgampwall = (TgK_wall * altmax - (Tstart_wall)) + (Tstart_wall) # Old
        Tgampwall = TgK_wall * altmax + Tstart_wall
        Tg = Tgamp * np.sin((((dectime - np.floor(dectime)) - SNUP / 24) / (TmaxLST / 24 - SNUP / 24)) * np.pi / 2) # 2015 a, based on max sun altitude
        Tgwall = Tgampwall * np.sin((((dectime - np.floor(dectime)) - SNUP / 24) / (TmaxLST_wall / 24 - SNUP / 24)) * np.pi / 2) # 2015a, based on max sun altitude

        if Tgwall < 0:  # temporary for removing low Tg during morning 20130205
            # Tg = 0
            Tgwall = 0

        # New estimation of Tg reduction for non - clear situation based on Reindl et al.1990
        radI0, _ = diffusefraction(I0, altitude, 1., Ta, RH)
        corr = 0.1473 * np.log(90 - (zen / np.pi * 180)) + 0.3454  # 20070329 correction of lat, Lindberg et al. 2008
        CI_Tg = (radG / radI0) + (1 - corr)
        if (CI_Tg > 1) or (CI_Tg == np.inf):
            CI_Tg = 1

        deg2rad = np.pi/180
        radG0 = radI0 * (np.sin(altitude * deg2rad)) + _
        CI_TgG = (radG / radG0) + (1 - corr)
        if (CI_TgG > 1) or (CI_TgG == np.inf):
            CI_TgG = 1
        
        # Tg = Tg * CI_Tg  # new estimation
        # Tgwall = Tgwall * CI_Tg
        Tg = Tg * CI_TgG  # new estimation
        Tgwall = Tgwall * CI_TgG
        if landcover == 1:
            Tg[Tg < 0] = 0  # temporary for removing low Tg during morning 20130205

        # # # # Ground View Factors # # # #
        gvfLup, gvfalb, gvfalbnosh, gvfLupE, gvfalbE, gvfalbnoshE, gvfLupS, gvfalbS, gvfalbnoshS, gvfLupW, gvfalbW,\
        gvfalbnoshW, gvfLupN, gvfalbN, gvfalbnoshN, gvfSum, gvfNorm = gvf_2018a(wallsun, walls, buildings, scale, shadow, first,
                second, dirwalls, Tg, Tgwall, Ta, emis_grid, ewall, alb_grid, SBC, albedo_b, rows, cols,
                                                                 Twater, lc_grid, landcover)

        # # # # Lup, daytime # # # #
        # Surface temperature wave delay - new as from 2014a
        Lup, timeaddnotused, Tgmap1 = TsWaveDelay_2015a(gvfLup, firstdaytime, timeadd, timestepdec, Tgmap1)
        LupE, timeaddnotused, Tgmap1E = TsWaveDelay_2015a(gvfLupE, firstdaytime, timeadd, timestepdec, Tgmap1E)
        LupS, timeaddnotused, Tgmap1S = TsWaveDelay_2015a(gvfLupS, firstdaytime, timeadd, timestepdec, Tgmap1S)
        LupW, timeaddnotused, Tgmap1W = TsWaveDelay_2015a(gvfLupW, firstdaytime, timeadd, timestepdec, Tgmap1W)
        LupN, timeaddnotused, Tgmap1N = TsWaveDelay_2015a(gvfLupN, firstdaytime, timeadd, timestepdec, Tgmap1N)
        
        # # For Tg output in POIs
        TgTemp = Tg * shadow + Ta
        TgOut, timeadd, TgOut1 = TsWaveDelay_2015a(TgTemp, firstdaytime, timeadd, timestepdec, TgOut1) #timeadd only here v2021a

        # Building height angle from svf
        F_sh = cylindric_wedge(zen, svfalfa, rows, cols)  # Fraction shadow on building walls based on sun alt and svf
        F_sh[np.isnan(F_sh)] = 0.5

        # # # # # # # Calculation of shortwave daytime radiative fluxes # # # # # # #
        Kdown = radI * shadow * np.sin(altitude * (np.pi / 180)) + dRad + albedo_b * (1 - svfbuveg) * \
                            (radG * (1 - F_sh) + radD * F_sh) # *sin(altitude(i) * (pi / 180))

        #Kdown = radI * shadow * np.sin(altitude * (np.pi / 180)) + radD * svfbuveg + albedo_b * (1 - svfbuveg) * \
                            #(radG * (1 - F_sh) + radD * F_sh) # *sin(altitude(i) * (pi / 180))

        Kup, KupE, KupS, KupW, KupN = Kup_veg_2015a(radI, radD, radG, altitude, svfbuveg, albedo_b, F_sh, gvfalb,
                    gvfalbE, gvfalbS, gvfalbW, gvfalbN, gvfalbnosh, gvfalbnoshE, gvfalbnoshS, gvfalbnoshW, gvfalbnoshN)

        #Keast, Ksouth, Kwest, Knorth, KsideI, KsideD = Kside_veg_v2019a(radI, radD, radG, shadow, svfS, svfW, svfN, svfE,
        #            svfEveg, svfSveg, svfWveg, svfNveg, azimuth, altitude, psi, t, albedo_b, F_sh, KupE, KupS, KupW,
        #            KupN, cyl, lv, ani, diffsh, rows, cols)

        Keast, Ksouth, Kwest, Knorth, KsideI, KsideD, Kside = Kside_veg_v2022a(radI, radD, radG, shadow, svfS, svfW, svfN, svfE,
                    svfEveg, svfSveg, svfWveg, svfNveg, azimuth, altitude, psi, t, albedo_b, F_sh, KupE, KupS, KupW,
                    KupN, cyl, lv, anisotropic_sky, diffsh, rows, cols, asvf, shmat, vegshmat, vbshvegshmat)

        firstdaytime = 0

    else:  # # # # # # # NIGHTTIME # # # # # # # #

        Tgwall = 0
        # CI_Tg = -999  # F_sh = []

        # Nocturnal K fluxes set to 0
        Knight = np.zeros((rows, cols))
        Kdown = np.zeros((rows, cols))
        Kwest = np.zeros((rows, cols))
        Kup = np.zeros((rows, cols))
        Keast = np.zeros((rows, cols))
        Ksouth = np.zeros((rows, cols))
        Knorth = np.zeros((rows, cols))
        KsideI = np.zeros((rows, cols))
        KsideD = np.zeros((rows, cols))
        F_sh = np.zeros((rows, cols))
        Tg = np.zeros((rows, cols))
        shadow = np.zeros((rows, cols))
        CI_Tg = deepcopy(CI)
        CI_TgG = deepcopy(CI)

        dRad = np.zeros((rows,cols))

        Kside = np.zeros((rows,cols))

        # # # # Lup # # # #
        Lup = SBC * emis_grid * ((Knight + Ta + Tg + 273.15) ** 4)
        if landcover == 1:
            Lup[lc_grid == 3] = SBC * 0.98 * (Twater + 273.15) ** 4  # nocturnal Water temp

        LupE = Lup
        LupS = Lup
        LupW = Lup
        LupN = Lup

        # # For Tg output in POIs
        TgOut = Ta + Tg

        I0 = 0
        timeadd = 0
        firstdaytime = 1

    # # # # Ldown # # # #
    # Anisotropic sky longwave radiation
    if anisotropic_sky == 1:
        if 'lv' not in locals():
            # Creating skyvault of patches of constant radians (Tregeneza and Sharples, 1993)
            skyvaultalt, skyvaultazi, _, _, _, _, _ = create_patches(patch_option)

            patch_emissivities = np.zeros(skyvaultalt.shape[0])

            x = np.transpose(np.atleast_2d(skyvaultalt))
            y = np.transpose(np.atleast_2d(skyvaultazi))
            z = np.transpose(np.atleast_2d(patch_emissivities))

            L_patches = np.append(np.append(x, y, axis=1), z, axis=1)

        else:
            L_patches = deepcopy(lv)

        if altitude < 0:
            CI = deepcopy(CI)

        if CI < 0.95:
            esky_c = CI * esky + (1 - CI) * 1.
            esky = esky_c

        Ldown, Lside, Least_, Lwest_, Lnorth_, Lsouth_ \
                  = Lcyl_v2022a(esky, L_patches, Ta, Tgwall, ewall, Lup, shmat, vegshmat, vbshvegshmat, 
                                altitude, azimuth, rows, cols, asvf)

    else:
        Ldown = (svf + svfveg - 1) * esky * SBC * ((Ta + 273.15) ** 4) + (2 - svfveg - svfaveg) * ewall * SBC * \
                    ((Ta + 273.15) ** 4) + (svfaveg - svf) * ewall * SBC * ((Ta + 273.15 + Tgwall) ** 4) + \
                    (2 - svf - svfveg) * (1 - ewall) * esky * SBC * ((Ta + 273.15) ** 4)  # Jonsson et al.(2006)
        # Ldown = Ldown - 25 # Shown by Jonsson et al.(2006) and Duarte et al.(2006)

        Lside = np.zeros((rows, cols))
        L_patches = None
 
        if CI < 0.95:  # non - clear conditions
            c = 1 - CI
            Ldown = Ldown * (1 - c) + c * ((svf + svfveg - 1) * SBC * ((Ta + 273.15) ** 4) + (2 - svfveg - svfaveg) *
                    ewall * SBC * ((Ta + 273.15) ** 4) + (svfaveg - svf) * ewall * SBC * ((Ta + 273.15 + Tgwall) ** 4) +
                    (2 - svf - svfveg) * (1 - ewall) * SBC * ((Ta + 273.15) ** 4))  # NOT REALLY TESTED!!! BUT MORE CORRECT?

    # # # # Lside # # # #
    Least, Lsouth, Lwest, Lnorth = Lside_veg_v2022a(svfS, svfW, svfN, svfE, svfEveg, svfSveg, svfWveg, svfNveg,
                    svfEaveg, svfSaveg, svfWaveg, svfNaveg, azimuth, altitude, Ta, Tgwall, SBC, ewall, Ldown,
                                                      esky, t, F_sh, CI, LupE, LupS, LupW, LupN, anisotropic_sky)

    # Box and anisotropic longwave
    if cyl == 0 and anisotropic_sky == 1:
        Least += Least_
        Lwest += Lwest_
        Lnorth += Lnorth_
        Lsouth += Lsouth_

    # # # # Calculation of radiant flux density and Tmrt # # # #
    # Human body considered as a cylinder with isotropic all-sky diffuse
    if cyl == 1 and anisotropic_sky == 0: 
        Sstr = absK * (KsideI * Fcyl + (Kdown + Kup) * Fup + (Knorth + Keast + Ksouth + Kwest) * Fside) + absL * \
                        ((Ldown + Lup) * Fup + (Lnorth + Least + Lsouth + Lwest) * Fside)
    # Human body considered as a cylinder with Perez et al. (1993) (anisotropic sky diffuse) 
    # and Martin and Berdahl (1984) (anisotropic sky longwave)
    elif cyl == 1 and anisotropic_sky == 1:
        Sstr = absK * (Kside * Fcyl + (Kdown + Kup) * Fup + (Knorth + Keast + Ksouth + Kwest) * Fside) + absL * \
                        ((Ldown + Lup) * Fup + Lside * Fcyl + (Lnorth + Least + Lsouth + Lwest) * Fside)
    # Knorth = nan Ksouth = nan Kwest = nan Keast = nan
    else: # Human body considered as a standing cube
        Sstr = absK * ((Kdown + Kup) * Fup + (Knorth + Keast + Ksouth + Kwest) * Fside) + absL * \
                        ((Ldown + Lup) * Fup + (Lnorth + Least + Lsouth + Lwest) * Fside)

    Tmrt = np.sqrt(np.sqrt((Sstr / (absL * SBC)))) - 273.2

    # Add longwave to cardinal directions for output in POI
    if (cyl == 1) and (anisotropic_sky == 1):
        Least += Least_
        Lwest += Lwest_
        Lnorth += Lnorth_
        Lsouth += Lsouth_

    return Tmrt, Kdown, Kup, Ldown, Lup, Tg, ea, esky, I0, CI, shadow, firstdaytime, timestepdec, \
           timeadd, Tgmap1, Tgmap1E, Tgmap1S, Tgmap1W, Tgmap1N, Keast, Ksouth, Kwest, Knorth, Least, \
           Lsouth, Lwest, Lnorth, KsideI, TgOut1, TgOut, radI, radD, \
               Lside, L_patches, CI_Tg, CI_TgG, KsideD, dRad, Kside
